import unittest
import numpy as _np
from mdciao.examples import filenames as test_filenames


from mdciao.utils import lists
import mdtraj as md

#TODO correct the "just works" nomenclature and the repetition of names of the funciton in class methods
class Test_exclude_same_fragments_from_residx_pairlist(unittest.TestCase):

    def test_exclude_same_fragments_from_residx_pairlist_just_works(self):
        assert (lists.exclude_same_fragments_from_residx_pairlist([[0, 1], [2, 3]], [[0, 1, 2], [3, 4]]) == [[2, 3]])

    def test_exclude_same_fragments_from_residx_pairlist_return_excluded_id(self):
        assert (lists.exclude_same_fragments_from_residx_pairlist([[1, 2], [0, 3], [5, 6]], [[0, 1, 2], [3, 4], [5, 6]],
                                                        return_excluded_idxs=True)
            == [0, 2])

class Test_unique_list_of_iterables_by_tuple_hashing(unittest.TestCase):

    def test_works(self):
        self.assertListEqual(lists.unique_list_of_iterables_by_tuple_hashing(
            [[0, 1], [1, 0], "ABC", [3, 4], [0, 1], "ABC", "DEF", [0, 1, 2]]),
            [[0, 1], [1, 0], "ABC", [3, 4],                "DEF", [0, 1, 2]])

    def test_works_array(self):
        [_np.testing.assert_array_equal(ii,jj) for ii, jj in zip(
            lists.unique_list_of_iterables_by_tuple_hashing(
            [[0, 1], [1, 0], "ABC", [3, 4], [0, 1], "ABC", "DEF", _np.array([0, 1, 2])]),
            [[0, 1], [1, 0], "ABC", [3, 4],                "DEF", [0, 1, 2]])]

    def test_returns_index(self):
        self.assertListEqual(lists.unique_list_of_iterables_by_tuple_hashing(
            [[0, 1], [1, 0], "ABC", [3, 4], [0, 1], "ABC", "DEF", [0, 1, 2]], return_idxs=True),
             [0, 1, 2, 3, 6,7])

    def test_works_for_non_iterables(self):
        self.assertListEqual(lists.unique_list_of_iterables_by_tuple_hashing([[1, 2], [3, 4], 1]),
                             [[1,2], [3,4],[1]])

    def test_bad_input(self):
        self.assertListEqual(["A"],lists.unique_list_of_iterables_by_tuple_hashing(["A"]))
        self.assertListEqual([0], lists.unique_list_of_iterables_by_tuple_hashing(["A"], return_idxs=True))

    def test_ignore_order_True(self):
        self.assertListEqual(lists.unique_list_of_iterables_by_tuple_hashing(
            [[0, 1], [1, 0], "ABC", [3, 4], [0, 1], "ABC", "DEF", [0, 1, 2]], ignore_order=True),
            [[0, 1],         "ABC", [3, 4],                "DEF", [0, 1, 2]])

class Test_in_what_fragment(unittest.TestCase):

    def test_works(self):
        # Easiest test
        assert lists.in_what_fragment(1, [[1, 2], [3, 4, 5, 6.6]]) == 0
        # Check that it returns the right fragments
        assert lists.in_what_fragment(1, [[1, 2], [3, 4, 5, 6.6]], ["A", "B"]) == 'A'

    def test_should_be_integer(self):
        # Check that it fails when input is not an index
        with self.assertRaises(AssertionError):
            lists.in_what_fragment([1],[[1, 2], [3, 4, 5, 6.6]])

        with self.assertRaises(AssertionError):
            lists.in_what_fragment([1], [[1, 2], [3, 4, 5, 6.6, "A"]])

class Test_does_not_contain_strings(unittest.TestCase):

    def test_works(self):
        assert lists.does_not_contain_strings([])
        assert lists.does_not_contain_strings([9, 99, 999])
        assert lists.does_not_contain_strings([9])
        assert not lists.does_not_contain_strings(["A"])
        assert not lists.does_not_contain_strings(["a", "b", "c"])
        assert not lists.does_not_contain_strings(["A", "b", "c"])
        assert not lists.does_not_contain_strings([[1], "ABC"])

class Test_force_iterable(unittest.TestCase):

    def test_works(self):
        assert len(lists.force_iterable("A")) != 0
        assert len(lists.force_iterable("_abc")) != 0
        assert len(lists.force_iterable([9, 99, 999])) != 0
        assert len(lists.force_iterable(999)) != 0

class Test_is_iterable(unittest.TestCase):

    def test_is_iterable(self):
        assert lists.is_iterable([]), lists.is_iterable([])
        assert lists.is_iterable("A")
        assert lists.is_iterable("_abc")
        assert lists.is_iterable([9, 99, 999])
        assert not lists.is_iterable(999)

class Test_in_what_N_fragments(unittest.TestCase):

    def setUp(self):
        self.fragments = [[0, 20, 30, 40],
                    [2, 4, 31, 1000],
                    [10]]

    def test_works(self):
        for fragment_idx, idx_query in enumerate([20, 4, 10]):
            assert len(lists.in_what_N_fragments(idx_query, self.fragments)) == 1

    def test_in_what_N_fragments_for_wrong_idx(self):
        wrong_idx = 11
        assert len(lists.in_what_N_fragments(wrong_idx, self.fragments)) == 1

class Test_rangeexpand(unittest.TestCase):
    def test_rangeexpand_just_works(self):
        assert (lists.rangeexpand("1-2, 3-4") == [1, 2, 3, 4])
        assert (lists.rangeexpand("1-2, 3,4") == [1, 2, 3, 4])
        assert (lists.rangeexpand("1-2, 03, 4") == [1, 2, 3, 4])

class Test_assert_min_len(unittest.TestCase):
    def test_assert_min_len_just_works(self):
        lists.assert_min_len([['a', 'b'], ['c', 'd'],[1, 2]])

    def test_assert_min_len_failed_assertion_just_works(self):
       with self.assertRaises(AssertionError):
           lists.assert_min_len([[1]])

    def test_assert_min_len_failed_assertion_works_empty_list(self):
        with self.assertRaises(AssertionError):
            lists.assert_min_len([[1,2],[]])

    def test_assert_min_length_min_len_works(self):
        lists.assert_min_len([['a']], min_len=1)

class Test_assert_no_intersection(unittest.TestCase):
    def test_assert_no_intersection_just_works(self):
        lists.assert_no_intersection([[1, 2], [3, 3]])

    def test_assert_no_intersection_empty_list_just_works(self):
        lists.assert_no_intersection([[], [3, 3]])

    def test_failed_assertion_just_works(self):
        with self.assertRaises(AssertionError):
            lists.assert_no_intersection([[1,2,3],[3,3]])

    def test_assert_no_intersection_empty_lists(self):
        with self.assertRaises(AssertionError):
            lists.assert_no_intersection([[], []])

class Test_window_average_fast(unittest.TestCase):
    def test_window_average_fast_just_works(self):
        assert _np.allclose(lists.window_average_fast(_np.arange(5)), _np.array([2.0]))
        assert _np.allclose(lists.window_average_fast(_np.arange(10)), _np.array([2.0, 3.0, 4.0, 5.0, 6.0, 7.0]))

    def test_window_average_fast_half_window_size_works(self):
        assert _np.allclose(lists.window_average_fast(_np.arange(7), half_window_size=3), _np.array([3.0]))
        assert _np.allclose(lists.window_average_fast(_np.arange(5), half_window_size=3), _np.array([1.42857143, 1.42857143, 1.42857143]))

class Test_join_lists(unittest.TestCase):
    def test_simple_run(self):
        in_lists = [[0, 1], [2, 3], [4, 5], [6, 7]]
        joined_lists = lists.join_lists(in_lists, [[1, 2]])
        assert _np.allclose(joined_lists[0], [0, 1])
        assert _np.allclose(joined_lists[1], [2, 3, 4, 5]), (joined_lists[1])
        assert _np.allclose(joined_lists[2], [6, 7]), (joined_lists[2])

    def test_re_ordered_run(self):
        in_lists = [[0, 1], [2, 3], [4, 5], [6, 7]]
        joined_lists = lists.join_lists(in_lists, [[3, 0],
                                             [1, 2]])
        assert _np.allclose(joined_lists[0], [0, 1, 6, 7])
        assert _np.allclose(joined_lists[1], [2, 3, 4, 5])


    def test_fails_overalpping_idxs(self):
        with self.assertRaises(AssertionError):
            in_lists = [[0, 1], [2, 3], [4, 5], [6, 7]]
            lists.join_lists(in_lists, [[1, 2],
                                  [2, 3]])

    def test_fails_no_minlen_list2join(self):
        with self.assertRaises(AssertionError):
            in_lists = [[0, 1], [2, 3], [4, 5], [6, 7]]
            lists.join_lists(in_lists, [[0],[1,2]])


class Test_put_this_idx_first_in_pair(unittest.TestCase):
    def test_works(self):
        assert(lists.put_this_idx_first_in_pair(20, [10,20]) == [20,10])
        assert (lists.put_this_idx_first_in_pair(10, [10, 20]) == [10, 20])
        assert (lists.put_this_idx_first_in_pair("first", ["first", "last"]) == ["first", "last"])
        assert (lists.put_this_idx_first_in_pair("first", ["last", "first"]) == ["first", "last"])

    def test_put_this_idx_first_in_pair_exception(self):
        with self.assertRaises(Exception):
            lists.put_this_idx_first_in_pair(99, [10, 20])

class Test_rewarp(unittest.TestCase):
    def test_int_input(self):
        assert all([_np.allclose(ii,jj) for ii,jj in zip([[0,1],
                                                          [2,3]],
                                                         lists.re_warp([0,1,2,3], 2))])

    def test_short_input(self):
        assert _np.allclose(lists.re_warp([0,1,2,3],[2]),[0,1])

class test_hash_list(unittest.TestCase):

    def test_passes(self):
        geom = md.load(test_filenames.small_monomer)
        ilist = [_np.random.randn(5000,2), # ndarray
                 [_np.random.randn(5000,2),_np.random.randn(5000,2)], #list thereof
                 'this_is_a_string', # str
                 ['this_is_a_substring', 'this_is_another_one'], # list thereof
                 1, # int
                 [1,2], #list of ints
                 [[1,],[2]], # list of lists,
                 geom,
                 geom.top,
                 [geom, geom],
                 [geom.top],
                 [geom, geom.top],
                 ]

        lists.hash_list(ilist)

class Test_contiguous_ranges(unittest.TestCase):

    def test_works(self):
        input = ["A","A", "A","0","A","A","0","0"]
        out_dict = lists.contiguous_ranges(input)
        print(out_dict)
        _np.testing.assert_array_equal(out_dict["A"][0],[0,1,2])
        _np.testing.assert_array_equal(out_dict["A"][1],[4,5])
        _np.testing.assert_equal(2,len(out_dict["A"]))
        _np.testing.assert_array_equal(out_dict["0"][0],[3])
        _np.testing.assert_array_equal(out_dict["0"][1],[6,7])
        _np.testing.assert_equal(2,len(out_dict["0"]))


class Test_idx_at_fraction(unittest.TestCase):

    def test_works(self):
        ncf = lists.idx_at_fraction([1, 1, 1], frac=2 / 3)
        _np.testing.assert_equal(ncf,1)

    def test_one_value(self):
        ncf = lists.idx_at_fraction([1], frac=2 / 3)
        _np.testing.assert_equal(ncf,0)

class Test_remove_from_lists(unittest.TestCase):
    def test_works(self):
        l = [
            [1, 2, 400, 6, 400, 6, 8],
            [400, 600, 700],
            [1, 2, 3, 4]
        ]
        nl = lists.remove_from_lists(l, [400,600,700])
        _np.testing.assert_array_equal(nl[0],[1,2,6,6,8])
        _np.testing.assert_array_equal(nl[1],[1,2,3,4])
    def test_weird(self):
        self.assertEqual([], lists.remove_from_lists([], [400, 600, 700]))
        self.assertEqual([], lists.remove_from_lists([[]], []))
        self.assertEqual([[400,500]], lists.remove_from_lists([[400,500]], []))

class Test_find_parent_list(unittest.TestCase):

    def test_works(self):
        fragments = [_np.arange(10),
                      _np.arange(50,100),
                       _np.arange(1000,2000)]
        subfragments = [[0,1,3], [4, 5, 6],
                        [10, 11, 12],
                        [60],
                        [1500,1600,1700],
                        [0]]
        parents_by_kid, kid_by_parents = lists.find_parent_list(subfragments, fragments)
        self.assertListEqual(parents_by_kid,[0,0, None,1,2,0])
        self.assertListEqual(list(kid_by_parents.keys()),[0,1,2])
        _np.testing.assert_array_equal(kid_by_parents[0],[0, 1, 5])
        _np.testing.assert_array_equal(kid_by_parents[1],[3])
        _np.testing.assert_array_equal(kid_by_parents[2],[4])

class Test_unique_product_w_intersection(unittest.TestCase):

    def test_works(self):
        a1 = [0,1,2,3]
        a2 = [2,3,4,5]

        res = lists.unique_product_w_intersection(a1,a2)

        self.assertListEqual(res.tolist(),
                                 [[0,2], #self a1-a1
                                  [0,3], #self a1-a1
                                  [0,4],
                                  [0,5],
                                  [1,2], #self a1-a1
                                  [1,3], #self a1-a1
                                  [1,4],
                                  [1,5],
                                  [2,3],
                                  [2,4],
                                  [2,5],
                                  [3,4],
                                  [3,5],
                                  ])

if __name__ == '__main__':
    unittest.main()




import unittest
import numpy as _np

from matplotlib import pyplot as _plt, colormaps as _cm
from matplotlib.colors import is_color_like, to_rgb

import mdciao.contacts
from mdciao.contacts import ContactGroup, ContactPair
from mdciao.examples import ContactGroupL394
from mdciao.cli import interface as _cli_interface
from mdciao import plots
from mdciao.plots.plots import _plot_freqbars_baseplot, _offset_dict, color_dict_guesser, _try_colormap_string, _plot_violin_baseplot, _color_tiler

from tempfile import TemporaryDirectory as _TDir
import os
import mdtraj as _md

from mdciao.examples import filenames as test_filenames

class TestPlotContactMatrix(unittest.TestCase):

    def test_plot_contact_matrix_runs_w_options(self):

        mat = _np.linspace(0,1,6).reshape(2,3)
        labels = [[1,2],["A","B","C"]]

        iax, ipix = plots.plot_matrix(mat,
                                      labels,
                                      grid=True,
                                      colorbar=True,
                                      transpose=True
                                      )
        #ax.figure.savefig("test.png",bbox_inches="tight")
        _plt.close("all")

    def test_plot_contact_matrix_raises_labels(self):

        mat = _np.linspace(0,1,6).reshape(2,3)
        labels = [[1,2],["A","B"]]
        with self.assertRaises(AssertionError):
            plots.plot_matrix(mat, labels)


class Test_plot_unified_freq_dicts(unittest.TestCase):

    def setUp(self):
        self.CG1_freqdict =          {"0-1":1,  "0-2":.75, "0-3":.50, "4-6":.25}
        self.CG1_freqdict_shuffled = {"0-2":.75, "0-1":1, "0-3":.50, "4-6":.25}

        self.CG2_freqdict = {"0-1":.95, "0-2":.65, "0-3":.35, "4-6":0}


    def test_plot_unified_freq_dicts_minimal(self):
        myfig, myax, __ = plots.plot_unified_freq_dicts({"CG1":self.CG1_freqdict, "CG1copy":self.CG1_freqdict},
                                                  {"CG1":"r", "CG1copy":"b"})

        #myfig.savefig("1.test_full.png", bbox_inches="tight")
        _plt.close("all")

    def test_plot_unified_freq_dicts_remove_identities(self):
        myfig, myax, __ = plots.plot_unified_freq_dicts({"CG1": self.CG1_freqdict, "CG1copy": self.CG1_freqdict},
                                                  {"CG1": "r", "CG1copy": "b"},
                                                  remove_identities=True)
        #Check that the 0-1 contact has been removed
        #myfig.savefig("2.test_wo_ident.png", bbox_inches="tight")
        _plt.close("all")

    def test_plot_unified_freq_dicts_remove_identities_cutoff(self):
        myfig, myax, __ = plots.plot_unified_freq_dicts({"CG1": self.CG1_freqdict, "CG2": self.CG2_freqdict},
                                                  {"CG1": "r", "CG2": "b"},
                                                  remove_identities=True)
        #Check that the 0-1 contact is there
        #myfig.savefig("3.test_wo_ident_cutoff.ref.png", bbox_inches="tight")
        _plt.close("all")

        myfig, myax, __ = plots.plot_unified_freq_dicts({"CG1": self.CG1_freqdict, "CG2": self.CG2_freqdict},
                                                  {"CG1": "r", "CG2": "b"},
                                                  remove_identities=True,
                                                  identity_cutoff=.95)
        #Check that the 0-1 contact has been removed
        #myfig.savefig("4.test_wo_ident_cutoff.png", bbox_inches="tight")
        _plt.close("all")

    def test_plot_unified_freq_dicts_lower_cutoff_val(self):
        myfig, myax, __ = plots.plot_unified_freq_dicts({"CG1": self.CG1_freqdict, "CG2": self.CG2_freqdict},
                                                  {"CG1": "r", "CG2": "b"},
                                                  remove_identities=True,
                                                  identity_cutoff=.9,
                                                  lower_cutoff_val=.4
                                                  )
        #myfig.savefig("5.test_above_below_thres.png", bbox_inches="tight")
        _plt.close("all")

        myfig, myax, __ = plots.plot_unified_freq_dicts({"CG1": self.CG1_freqdict, "CG2": self.CG2_freqdict},
                                                  {"CG1": "r", "CG2": "b"},
                                                  remove_identities=True,
                                                  identity_cutoff=.95)
        #myfig.savefig("6.test_wo_ident_cutoff.png", bbox_inches="tight")
        _plt.close("all")


    def test_plot_unified_freq_dicts_order_std(self):
        myfig, myax, __ = plots.plot_unified_freq_dicts({"CG1": self.CG1_freqdict, "CG2": self.CG2_freqdict},
                                                  {"CG1": "r", "CG2": "b"},
                                                  sort_by="std",
                                                  lower_cutoff_val=0.0
                                                  )
        # Order should be inverted
        #myfig.savefig("7.test_order.std.png", bbox_inches="tight")
        _plt.close("all")

    def test_plot_unified_freq_dicts_order_keep(self):
        myfig, myax, __ = plots.plot_unified_freq_dicts({"CG1s": self.CG1_freqdict_shuffled, "CG2": self.CG2_freqdict},
                                                  {"CG1s": "r", "CG2": "b"},
                                                  sort_by="keep",
                                                  )
        # myfig.savefig("8.test_order.keep.png", bbox_inches="tight")
        _plt.close("all")




    def test_plot_unified_freq_dicts_remove_identities_vert(self):
        myfig, myax, __ = plots.plot_unified_freq_dicts({"CG1": self.CG1_freqdict, "CG2": self.CG2_freqdict},
                                                  {"CG1": "r", "CG2": "b"},
                                                  remove_identities=True,
                                                  vertical_plot=True)
        #myfig.savefig("9.test_wo_ident_vert.png", bbox_inches="tight")
        _plt.close("all")

    def test_plot_unified_freq_dicts_remove_identities_vert_order_by_std(self):
        myfig, myax, __ = plots.plot_unified_freq_dicts({"CG1": self.CG1_freqdict, "CG2": self.CG2_freqdict},
                                                  {"CG1": "r", "CG2": "b"},
                                                  sort_by="std",
                                                  #remove_identities=True,
                                                  lower_cutoff_val=0.0,
                                                  vertical_plot=True)
        #myfig.savefig("10.test_vert.std.png", bbox_inches="tight")
        _plt.close("all")

    def test_plot_unified_freq_dicts_ax(self):
        _plt.figure()
        ax = _plt.gca()
        myfig, myax, __ = plots.plot_unified_freq_dicts({"CG1":self.CG1_freqdict, "CG1copy":self.CG1_freqdict},
                                                  {"CG1":"r", "CG1copy":"b"}, ax=ax)
        assert myax is ax
        _plt.close("all")

    def test_plot_unified_freq_dictsfigsize_None(self):
        myfig, myax, __ = plots.plot_unified_freq_dicts({"CG1": self.CG1_freqdict, "CG1copy": self.CG1_freqdict},
                                                        {"CG1": "r", "CG1copy": "b"},
                                                        figsize=None)
        _plt.close("all")

    def test_plot_unified_freq_dictswinner(self):
        myfig, myax, __ = plots.plot_unified_freq_dicts({"CG1": {"0-1":0, "0-2":1},
                                                         "CG12": {"0-1":1, "0-2":0}},
                                                        figsize=None,
                                                        assign_w_color=True)
        _plt.close("all")

    def test_plot_just_one_dict(self):
        _plt.figure()
        ax = _plt.gca()
        myfig, myax, __ = plots.plot_unified_freq_dicts({"CG1": self.CG1_freqdict},
                                                  {"CG1": "r"})
        #myfig.savefig("12.test_just_one.png",bbox_inches="tight")
        _plt.close("all")

    def test_sort_by_keys(self):
        neigh : mdciao.contacts.ContactGroup = ContactGroupL394()
        freqs = neigh.frequency_dicts(4)
        # The following list, test_list
        # * reverses the order
        # * deletes the most frequent one (the last one) "R389@G.H5.21    - L394@G.H5.26"
        # * adds a bogus-key
        # * duplicates an entry "L388@G.H5.20    - L394@G.H5.26"
        test_list = list(freqs.keys())[::-1][:-1]+["bogus-key"]+["L388@G.H5.20    - L394@G.H5.26"]
        myfig, myax, plotted_freqs = plots.plot_unified_freq_dicts({"L394":freqs},
                                                        sort_by=test_list,
                                                        lower_cutoff_val=.5,
                                                        )
        #myfig.tight_layout()
        #myfig.savefig("test.keys.png")
        self.assertListEqual(test_list[1:-2], # First one will be missed bc.
                                              # lower_cutoff_val, last two because bogus and repetition
                             list(plotted_freqs["L394"].keys()))
        assert "L394" in plotted_freqs.keys()
        _plt.close("all")

    def test_sort_by_numeric(self):
        myfig, myax, plotted_freqs = plots.plot_unified_freq_dicts({"A": {"3-1": 0.1, "2-1": .9}},
                                                                   sort_by="numeric"
                                                                   )
        #myfig.tight_layout()
        #myfig.savefig("test.keys.png")
        self.assertDictEqual(plotted_freqs["A"], {"2-1": 0.9, "3-1": .1})
        _plt.close("all")

    def test_sort_by_numeric_raises(self):
        with self.assertRaises(ValueError):
            plots.plot_unified_freq_dicts({"A": {"B-A": 0.1, "C-A": .9}},
                                          sort_by="numeric"
                                          )
    def test_sort_by_wrong_raises(self):
        with self.assertRaises(ValueError):
            plots.plot_unified_freq_dicts({"A": {"B-A": 0.1, "C-A": .9}},
                                          sort_by="numerica"
                                          )

class Test_pop_keys_by_scheme(unittest.TestCase):
    def setUp(self):
        self.CG1_freqdict = {"4-6": .25, "0-1": 1., "0-2": .75, "0-3": .50}
        self.CG2_freqdict = {"0-1": .99, "0-2": .50, "0-3": .1, "4-6": 0}

        self.unified_dict = {"CG1": self.CG1_freqdict, "CG2": self.CG2_freqdict}
        """
        unified_dict
              CG1   CG2
        4-6  0.25  0.00
        0-1  1.00  0.99
        0-2  0.75  0.50
        0-3  0.50  0.10
        
        dict_for_sorting
              mean    std 
        4-6  0.125  0.125 
        0-1  0.995  0.005 
        0-2  0.625  0.125 
        0-3  0.300  0.200 
        """
        self.dict_for_sorting = {"mean" : {}, "std" :  {}}
        for key in self.unified_dict["CG1"].keys():
            self.dict_for_sorting["std"][key]  = _np.std([idict[key] for idict in self.unified_dict.values()])
            self.dict_for_sorting["mean"][key] = _np.mean([idict[key] for idict in self.unified_dict.values()])

    def test_pop_keys_by_scheme_mean_remove_identities_and_low(self):
        all_ctc_keys, freqs_by_sys_by_ctc, ctc_keys_popped_above, ctc_keys_popped_below = \
            plots.plots._pop_keys_by_scheme("mean", self.unified_dict, self.dict_for_sorting,
                                        0.3, .95, True)

        self.assertListEqual(all_ctc_keys, ["0-2", "0-3"])
        self.assertListEqual(ctc_keys_popped_above,["0-1"])  # identity_cutoff = .95 drops 0-1
        self.assertListEqual(ctc_keys_popped_below,["4-6"])  # lower_cutoff_val = .3 drops 4-6

    def test_pop_keys_by_scheme_std_remove_identities_and_low(self):
        all_ctc_keys, freqs_by_sys_by_ctc, ctc_keys_popped_above, ctc_keys_popped_below = \
            plots.plots._pop_keys_by_scheme("std", self.unified_dict, self.dict_for_sorting,
                                        0.15, .95, False)

        self.assertListEqual(all_ctc_keys, ["0-3"])
        self.assertListEqual(ctc_keys_popped_above,[])  # we're not removing identity
        self.assertListEqual(ctc_keys_popped_below,["4-6", "0-1","0-2"])  # lower_cutoff_val = .15 drops these

    def test_pop_keys_by_scheme_list_remove_identities_and_low(self):
        all_ctc_keys, freqs_by_sys_by_ctc, ctc_keys_popped_above, ctc_keys_popped_below = \
            plots.plots._pop_keys_by_scheme("list", self.unified_dict, self.dict_for_sorting,
                                            0.15, .95, True)

        self.assertListEqual(all_ctc_keys, ["4-6", "0-2", "0-3"])
        self.assertListEqual(ctc_keys_popped_above,["0-1"])  # identity_cutoff = .95 drops 0-1
        self.assertListEqual(ctc_keys_popped_below, [])  # these aren't on the list

    def test_pop_keys_by_scheme_numeric_remove_identities_and_low(self):
        all_ctc_keys, freqs_by_sys_by_ctc, ctc_keys_popped_above, ctc_keys_popped_below = \
            plots.plots._pop_keys_by_scheme("numeric", self.unified_dict, self.dict_for_sorting,
                                            0.3, .95, True)

        self.assertListEqual(all_ctc_keys, ["0-2", "0-3"])
        self.assertListEqual(ctc_keys_popped_above, ["0-1"])  # identity_cutoff = .95 drops 0-1
        self.assertListEqual(ctc_keys_popped_below, ["4-6"])  # lower_cutoff_val = .3 drops 4-6

    def test_pop_keys_by_scheme_keep_remove_identities_and_low(self):
        all_ctc_keys, freqs_by_sys_by_ctc, ctc_keys_popped_above, ctc_keys_popped_below = \
            plots.plots._pop_keys_by_scheme("keep", self.unified_dict, self.dict_for_sorting,
                                            0.3, .95, True)

        self.assertListEqual(all_ctc_keys, ["0-2", "0-3"])
        self.assertListEqual(ctc_keys_popped_above, ["0-1"])  # identity_cutoff = .95 drops 0-1
        self.assertListEqual(ctc_keys_popped_below, ["4-6"])  # lower_cutoff_val = .3 drops 4-6

class Test_plot_unified_distro_dicts(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.CGL394 = ContactGroupL394()
        D1 = cls.CGL394.distribution_dicts()
        D2 = {key: [val[0],val[1]+.02] for key,val in D1.items()}
        cls.dicts = {"L394":D1,
                     "L394shifted":D2}

    def test_plot_unified_distro_dicts(self):
        plots.plot_unified_distro_dicts(self.dicts, ctc_cutoff_Ang=3.5, n_cols=2,legend_rows=1)
        #_plt.savefig("test.png")
        _plt.close("all")


    def test_plot_unified_distro_dicts_preexising_ax(self):
        myfig, myax = _plt.subplots(10,2)
        plots.plot_unified_distro_dicts(self.dicts, ctc_cutoff_Ang=3.5, n_cols=2,legend_rows=1, ax_array=myax)
        #_plt.savefig("test.png")
        _plt.close("all")

class Test_compare_groups_of_contacts(unittest.TestCase):

    def setUp(self):
        self.CG1 = ContactGroup([ContactPair([0,1],[[.1, .1]], [[0., 1.]]),
                                 ContactPair([0,2],[[.1, .2]], [[0., 1.]])])

        self.CG2 = ContactGroup([ContactPair([0,1],[[.1, .1, .1]], [[0., 1., 2.]]),
                                 ContactPair([0,3],[[.1, .2, .2]], [[0., 1., 2.]])])

    def test_just_works(self):
        myfig, freqs, __ = plots.compare_groups_of_contacts({"CG1":self.CG1,"CG2":self.CG2},
                                                      ctc_cutoff_Ang=1.5)
        myfig.tight_layout()
        #myfig.savefig("1.test.png",bbox_inches="tight")
        _plt.close("all")

    def test_just_works_list(self):
        myfig, freqs, __ = plots.compare_groups_of_contacts([self.CG1, self.CG2],
                                                            ctc_cutoff_Ang=1.5)
        myfig.tight_layout()
        #myfig.savefig("1.1.test.png",bbox_inches="tight")
        _plt.close("all")

    def test_per_residue(self):
        myfig, freqs, __ = plots.compare_groups_of_contacts([self.CG1, self.CG2],
                                                            ctc_cutoff_Ang=1.5,
                                                            per_residue=True)
        myfig.tight_layout()
        #myfig.savefig("1.per_residue.test.png",bbox_inches="tight")
        _plt.close("all")


    def test_mutation(self):
        myfig, freqs, __ = plots.compare_groups_of_contacts({"CG1":self.CG1,"CG2":self.CG2},
                                                      ctc_cutoff_Ang=1.5,
                                                      mutations_dict={"3":"2"})
        myfig.tight_layout()
        #myfig.savefig("2.test.png",bbox_inches="tight")
        _plt.close("all")

    def test_anchor(self):
        myfig, freqs, __ = plots.compare_groups_of_contacts({"CG1":self.CG1,"CG2":self.CG2},
                                                            ctc_cutoff_Ang=1.5,
                                                      anchor="0")
        myfig.tight_layout()
        #myfig.savefig("3.test.png",bbox_inches="tight")
        _plt.close("all")

    def test_plot_singles(self):
        myfig, freqs, __ = plots.compare_groups_of_contacts({"CG1":self.CG1,"CG2":self.CG2},
                                                      ctc_cutoff_Ang=1.5,
                                                      plot_singles=True)
        myfig.tight_layout()
        #myfig.savefig("4.test.png",bbox_inches="tight")
        _plt.close("all")

    def test_plot_singles_w_anchor(self):
        myfig, freqs, __ = plots.compare_groups_of_contacts({"CG1":self.CG1,"CG2":self.CG2},
                                                      ctc_cutoff_Ang=1.5,
                                                      anchor="0",
                                                      plot_singles=True)
        myfig.tight_layout()
        #myfig.savefig("5.test.png",bbox_inches="tight")
        _plt.close("all")

    def test_with_freqdat(self):
        with _TDir(suffix="_test_mdciao") as tmpdir:
            freqfile = os.path.join(tmpdir,"freqtest.dat")
            self.CG2.frequency_table(1.5, freqfile)
            myfig, __, __ = plots.compare_groups_of_contacts({"CG1":self.CG1,
                                                        "CG2":self.CG2,
                                                        "CG2f":freqfile},
                                                       {"CG1":"r", "CG2":"b", "CG2f":"g"},
                                                       ctc_cutoff_Ang=1.5)
            # myfig.savefig("6.test.png",bbox_inches="tight")
        _plt.close("all")

    def test_with_freqdict(self):
        myfig, __, __ = plots.compare_groups_of_contacts({"CG1":self.CG1,
                                                    "CG2":self.CG2,
                                                    "CG2d":{"0-1":1, "0-3":1/3}},
                                                   {"CG1":"r", "CG2":"b", "CG2d":"g"},
                                                   ctc_cutoff_Ang=1.5)
        # myfig.savefig("7.test.png",bbox_inches="tight")
        _plt.close("all")

    def test_color_list(self):
        myfig, freqs, __ = plots.compare_groups_of_contacts({"CG1": self.CG1, "CG2": self.CG2},
                                                            ["r","g"],
                                                            ctc_cutoff_Ang=1.5)
        myfig.tight_layout()
        # myfig.savefig("1.test.png",bbox_inches="tight")
        _plt.close("all")

    def test_with_CG_freqdict_file_as_list(self):
        with _TDir(suffix="_test_mdciao") as tmpdir:
            freqfile = os.path.join(tmpdir, "freqtest.dat")
            self.CG2.frequency_table(1.5,freqfile)
            myfig, __, __ = plots.compare_groups_of_contacts([self.CG1,
                                                              self.CG2,
                                                              {"0-1":1, "0-3":1/3},
                                                              freqfile],
                                                             ctc_cutoff_Ang=1.5)
            #myfig.savefig("8.test.png")
            _plt.close("all")

    def test_distro(self):
        CG : ContactGroup = ContactGroupL394()
        plots.compare_groups_of_contacts([CG,CG],distro=True)

    def test_interface(self):
        intf1 = _cli_interface(_md.load(test_filenames.actor_pdb),
                                fragments=[_np.arange(868, 875 + 1),  # ICL2
                                            _np.arange(328, 353 + 1)],  # Ga5,
                                ctc_cutoff_Ang=30,
                                no_disk=True,
                                figures=False
                                )
        intf2 = _cli_interface(_md.load(test_filenames.actor_pdb),
                                fragments=[_np.arange(860, 875 + 1),  # ICL2
                                            _np.arange(320, 353 + 1)],  # Ga5,
                                ctc_cutoff_Ang=30,
                                no_disk=True,
                                figures=False
                                )
        ifig, __, __  = plots.compare_groups_of_contacts({"1": intf1, "2":intf2}, ctc_cutoff_Ang=30,interface=True)
        #ifig.savefig("test.png")

class test_histogram_w_smoothing_auto(unittest.TestCase):

    def test_just_runs_all(self):
        ax = plots.histogram_w_smoothing_auto(_np.random.randn(200), bins=20, maxcount=3.14)
        for line in ax.lines:
            y = line.get_xydata()[:,1]
            assert y.max()==3.14


class Test_plot_w_smoothing_auto(unittest.TestCase):

    def test_just_runs(self):
        x = [0,1,2,3,4,5]
        y = [0,1,2,3,4,5]
        _plt.figure()
        line2D = plots.plot_w_smoothing_auto(y, _plt.gca(), "test", "r", x=x, background=True, n_smooth_hw=1)
        assert isinstance(line2D, _plt.Line2D)
        _plt.close(_plt.gcf())

    def test_creates_x(self):
        y = [0,1,2,3,4,5]
        _plt.figure()
        line2D = plots.plot_w_smoothing_auto(y, _plt.gca(), "test", "r", background=True, n_smooth_hw=1)
        assert isinstance(line2D, _plt.Line2D)
        _plt.close(_plt.gcf())

    def test_background_False(self):
        y = [0,1,2,3,4,5]
        _plt.figure()
        line2D = plots.plot_w_smoothing_auto(y, _plt.gca(), "test", "r", background=False, n_smooth_hw=1)
        assert isinstance(line2D, _plt.Line2D)
        _plt.close(_plt.gcf())

    def test_background_str(self):
        y = [0,1,2,3,4,5]
        _plt.figure()
        line2D = plots.plot_w_smoothing_auto(y, _plt.gca(), "test", "r", background="green", n_smooth_hw=1)
        assert isinstance(line2D, _plt.Line2D)
        _plt.close(_plt.gcf())

    def test_background_rgb(self):
        y = [0,1,2,3,4,5]
        _plt.figure()
        line2D = plots.plot_w_smoothing_auto(y, _plt.gca(), "test", "r", background=[1., 0., 0.], n_smooth_hw=1)
        assert isinstance(line2D, _plt.Line2D)
        _plt.close(_plt.gcf())

    def test_background_raises(self):
        y = [0,1,2,3,4,5]
        _plt.figure()
        with self.assertRaises(AssertionError):
            line2D = plots.plot_w_smoothing_auto(y, _plt.gca(), "test", "r", background=["A", 0., 0.], n_smooth_hw=1)

    def test_no_smoothing(self):
        y = [0,1,2,3,4,5]
        _plt.figure()
        line2D = plots.plot_w_smoothing_auto(y, _plt.gca(), "test", "r")
        assert isinstance(line2D, _plt.Line2D)
        _plt.close(_plt.gcf())

    def test_plot_new_axis(self):
        y = [0, 1, 2, 3, 4, 5]
        line2D = plots.plot_w_smoothing_auto(y, label= "test", color="r", n_smooth_hw=1)
        assert isinstance(line2D, _plt.Line2D)
        _plt.close(_plt.gcf())

    def test_plot_existing_axis(self):
        y = [0, 1, 2, 3, 4, 5]
        _plt.plot(y)
        _plt.text(0,0,"this axes existed before")
        iax = _plt.gca()
        n_lines_before = len(iax.lines)
        line2D = plots.plot_w_smoothing_auto(y, label= "test", color="r", n_smooth_hw=1)
        assert isinstance(line2D, _plt.Line2D)
        assert "this axes existed before" in [txt.get_text() for txt in line2D.axes.texts]
        assert len(line2D.axes.lines)==n_lines_before +2 #original lines plus two from the smoothing
        assert iax is line2D.axes
        _plt.close(_plt.gcf())




class Test_color_by_values(unittest.TestCase):
    def setUp(self):
        self.freqs_by_sys_by_ctc = {"WT":   {"A": 1, "B": .0,  "C": 1.0},
                                    "mut1": {"A": 0, "B": .0,  "C": .25},
                                    "mut2": {"A": 1, "B": .75, "C": .25}}
        self.colordict = {"WT": "r", "mut1": "b", "mut2": "g"}
        from mdciao.plots import plots as _plots
        self.plots = _plots


    def test_runs(self):
        info = self.plots._color_by_values(["A", "B", "C"], self.freqs_by_sys_by_ctc,
                                           self.colordict)
        _np.testing.assert_equal(info,
                                 {"A": ("-", self.colordict["mut1"]),
                                  # the label gets the color of mut1 (only one winner)
                                  "B": ("+", self.colordict["mut2"]),
                                  # the label gets the color of mut2 (only one loser)
                                  "C": ("", "k")})  # nothing happens, neither winners nor losers

    def test_runs_wo_action(self):
        info = self.plots._color_by_values(["A", "B", "C"], self.freqs_by_sys_by_ctc,
                                           self.colordict,
                                           assign_w_color=False,
                                           )

        # nothing happens
        _np.testing.assert_equal(info,
                                 {"A": ("", "k"),
                                  "B": ("", "k"),
                                  "C": ("", "k")})

class Test_freqs_baseplot(unittest.TestCase):

    def test_baseplot_minimal(self):
        jax = _plot_freqbars_baseplot([1,2,3])
        assert isinstance(jax, _plt.Axes)
        _plt.close("all")

    def test_baseplot_pass_ax(self):
        _plt.plot()
        jax = _plt.gca()
        assert jax is _plot_freqbars_baseplot([1,2,3], ax=jax)
        _plt.close("all")

    def test_baseplot_truncate(self):

        jax = _plot_freqbars_baseplot([1,2,3], lower_cutoff_val=.5)

        _plt.close("all")
if __name__ == '__main__':
    unittest.main()


class Test_offset_dict(unittest.TestCase):

    def test_one(self):
        # Only one bar, padding is .2, hence position is 0 and width is 1-.2 = .8
        res, width = _offset_dict(["A"])
        self.assertDictEqual(res,{"A":0})
        self.assertEqual(width,.8)

    def test_two(self):

        res, width = _offset_dict(["left","right"])

        # Twice as many bars, with same padding, i.e. width = .8/2 = .4
        self.assertEqual(width,.4)

        # If width is .4 and there's two of them, they have to be centered around -.2 and +.2
        self.assertDictEqual(res,{"left":  -.2,
                                  "right":  .2})

    def test_three(self):
        res, width = _offset_dict(["left", "middle", "right"], width=.2)


        # Three with fixed width .2 means -.2, 0, +.2
        self.assertDictEqual(res, {"left": -.2,
                                   "middle":0,
                                   "right": +.2})

        self.assertEqual(width,.2)

class Test_sorter_by_key_or_val(unittest.TestCase):

    def setUp(self):
        self.indict = {'R131-E392': 4.1,
                       'P138-M386': 6.85,
                       'K270-L394': 4.68,
                       'T274-L393': 4.95,
                       'Q142-I382': 7.73,
                       'T68-R389': 7.77,
                       'Y141-R389': 8.02,
                       'T274-E392': 5.73,
                       'V222-L393': 6.92}

    def test_residue(self):
        sorted_keys = plots.plots._sorter_by_key_or_val("residue", self.indict)

        self.assertListEqual(sorted_keys, [
            'T68-R389',
            "R131-E392",
            "P138-M386",
            'Y141-R389',
            'Q142-I382',
            'V222-L393',
            'K270-L394',
            'T274-E392',
            'T274-L393',
        ])

    def test_mean(self):
        sorted_keys = plots.plots._sorter_by_key_or_val("mean", self.indict)

        self.assertListEqual(sorted_keys,
                             list({'R131-E392': 4.1,
                                   'K270-L394': 4.68,
                                   'T274-L393': 4.95,
                                   'T274-E392': 5.73,
                                   'P138-M386': 6.85,
                                   'V222-L393': 6.92,
                                   'Q142-I382': 7.73,
                                   'T68-R389': 7.77,
                                   'Y141-R389': 8.02,
                                   }.keys()))

    def test_some_non_numeric(self):
        sorted_keys = plots.plots._sorter_by_key_or_val("numeric", {key : None for key in ["0-20", "0-10", "ALA30-GLU50", "ALA30-GLU40", "ALA", "GLU5-ALA20"]})
        self.assertListEqual(sorted_keys, ['0-10', '0-20', 'GLU5-ALA20', 'ALA30-GLU40', 'ALA30-GLU50', 'ALA'])
    def test_keep(self):
        sorted_keys = plots.plots._sorter_by_key_or_val("keep", {key : None for key in ["0-20", "0-10", "ALA30-GLU50", "ALA30-GLU40", "ALA", "GLU5-ALA20"]})
        self.assertListEqual(sorted_keys, ["0-20", "0-10", "ALA30-GLU50", "ALA30-GLU40", "ALA", "GLU5-ALA20"])

class Test_colormaps(unittest.TestCase):

    def test_works(self):
        colors = _try_colormap_string("jet", 10)
        self.assertEqual(len(colors),10)
        assert all([is_color_like(col) for col in colors])

    def test_works_large_N(self):
        colors = _try_colormap_string("tab10", 20)
        self.assertEqual(len(colors),20)
        assert all([is_color_like(col) for col in colors])
        _np.testing.assert_array_equal(colors[0:10], colors[10:20])

    def test_exception_raises(self):
        with self.assertRaises(AttributeError):
            _try_colormap_string("Chet",10)

    def test_color_dict_guesser_None(self):
        colors = color_dict_guesser(None, ["sys1", "sys2", "sys3"])
        colors_rgb_array = _np.vstack([to_rgb(col) for key, col in colors.items()])
        ref_color = _np.array(list(_cm["tab10"]([0,1,2])))[:,:-1]
        _np.testing.assert_array_equal(colors_rgb_array, ref_color)
        self.assertListEqual(list(colors.keys()), ["sys1","sys2","sys3"])

    def test_color_dict_guesser_cmap(self):
        colors = color_dict_guesser("Set2", ["sys1", "sys2", "sys3"])
        colors_rgb_array = _np.vstack([to_rgb(col) for key, col in colors.items()])
        ref_color = _np.array(list(_cm["Set2"]([0, 1, 2])))[:, :-1]
        _np.testing.assert_array_equal(colors_rgb_array, ref_color)
        self.assertListEqual(list(colors.keys()), ["sys1", "sys2", "sys3"])

    def test_color_dict_guesser_cmap_n(self):
        colors = color_dict_guesser("Set2", 3)
        colors_rgb_array = _np.vstack([to_rgb(col) for key, col in colors.items()])
        ref_color = _np.array(list(_cm["Set2"]([0, 1, 2])))[:, :-1]
        _np.testing.assert_array_equal(colors_rgb_array, ref_color)
        self.assertListEqual(list(colors.keys()), [0, 1, 2])

    def test_color_dict_guesser_list(self):
        colors = color_dict_guesser(["r", "g", "b"], ["sys1", "sys2"])
        self.assertDictEqual(colors, {"sys1":"r","sys2":"g"})

    def test_color_dict_guesser_dict(self):
        colors = color_dict_guesser({"sys1": "r", "sys2": "g"}, ["sys1", "sys2"])
        self.assertDictEqual(colors, {"sys1": "r", "sys2": "g"})

class Test_color_tiler(unittest.TestCase):

    def test_str(self):
        self.assertListEqual(['red', 'red', 'red', 'red', 'red'], _color_tiler("red", 5))  # pure str

    def test_list(self):
        self.assertListEqual(['red', 'red', 'red', 'red', 'red'],
                             _color_tiler(["red"], 5))  # list of str with 1 item

        self.assertListEqual(['red', 'blue', 'red', 'blue', 'red'],
                             _color_tiler(["red", "blue"], 5))  # list of strs

    def test_array(self):
        self.assertListEqual([[1.0, 0.0, 0.0],
                              [0.0, 1.0, 0.0],
                              [0.0, 0.0, 1.0],
                              [1.0, 0.0, 0.0],
                              [0.0, 1.0, 0.0]],
                             _color_tiler([[1., 0., 0.],  # red
                                           [0., 1., 0.],  # green
                                           [0., 0., 1.]], 5))  # blue

class Test_plot_violin_baseplot(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.CGL394 : ContactGroup = ContactGroupL394()
        cp: ContactPair

        cls.time_trajs = [cp.time_traces.ctc_trajs for cp in cls.CGL394._contacts]
        cls.time_trajs_shifted = [[itraj+.25 for itraj in cp] for cp in cls.time_trajs]
    def test_minimal(self):
        iax, _ = _plot_violin_baseplot(self.time_trajs, labels=self.CGL394.ctc_labels)
        #ax.figure.savefig("test.pdf")
        _plt.close("all")

    def test_second_axis(self):
        iax, _ = _plot_violin_baseplot(self.time_trajs)
        iax, _ = _plot_violin_baseplot(self.time_trajs_shifted, labels=self.CGL394.ctc_labels, ax=iax)
        #ax.figure.savefig("test.pdf")
        _plt.close("all")

class Test_plot_compare_violins(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.CGL394 : ContactGroup = ContactGroupL394()
        cls.CGL394_larger  = ContactGroupL394(ctc_control=.99, ctc_cutoff_Ang=5)

    def test_works(self):
        fig, ax, sorted_keys = plots.compare_violins({"small": self.CGL394, "big": self.CGL394_larger},
                                                     anchor="L394",
                                                     ymax=10, ctc_cutoff_Ang=4)
        # fig.savefig("test.pdf")
        _plt.close("all")

    def test_works_no_defrag_and_list_zero_freq_remove_identities(self):
        fig, ax, sorted_keys = plots.compare_violins([self.CGL394, self.CGL394_larger],
                                                     anchor="L394",
                                                     ymax=10, ctc_cutoff_Ang=4,
                                                     zero_freq=.5,
                                                     remove_identities=True,
                                                     identity_cutoff=.90,
                                                     defrag=None)
        # This should hold
        """
        {'L394@G.H5.26-I233@5.72x72': 0.39285714285714285, # below zero_freq
        'R389@G.H5.21-L394@G.H5.26': 0.9464285714285714,  # above identity
        'L388@G.H5.20-L394@G.H5.26': 0.9464285714285714,  # above identity
        'R385@G.H5.17-L394@G.H5.26': 0.6964285714285714, 
        'L394@G.H5.26-L230@5.69x69': 0.8392857142857143, 
        'L394@G.H5.26-K270@6.32x32': 0.4642857142857143,  # below zero_freq
        'L394@G.H5.26-K267@6.29x29': 0.0, # below zero_freq
        'L394@G.H5.26-Q229@5.68x68': 0.0} # below zero_freq
        """
        self.assertListEqual(sorted_keys, ["L230@5.69x69", 'R385@G.H5.17'])
        _plt.close("all")

    def test_works_no_defrag_and_list(self):
        fig, ax, sorted_keys = plots.compare_violins([self.CGL394, self.CGL394_larger],
                                                     anchor="L394",
                                                     ymax=10, ctc_cutoff_Ang=4,
                                                     defrag=None)

        #fig.savefig("test.pdf")
        _plt.close("all")

    def test_repframes_True(self):
        fig, ax, sorted_keys, repframes = plots.compare_violins({"small": self.CGL394, "big": self.CGL394_larger},
                                                                anchor="L394",
                                                                ymax=10, ctc_cutoff_Ang=4,
                                                                representatives=True)

        assert repframes["small"] is None
        assert repframes["big"] is None
        #fig.savefig("test.pdf")
        _plt.close("all")

    def test_repframes_int(self):
        fig, ax, sorted_keys, repframes = plots.compare_violins({"small": self.CGL394, "big": self.CGL394_larger},
                                                                anchor="L394",
                                                                ymax=10, ctc_cutoff_Ang=4,
                                                                representatives=2)
        assert repframes["small"] is None
        assert repframes["big"] is None
        # fig.savefig("test.pdf")
        _plt.close("all")

    def test_repframes_dict_kwargs(self):
        fig, ax, sorted_keys, repframes_out = plots.compare_violins({"small": self.CGL394, "big": self.CGL394_larger},
                                                                    anchor="L394",
                                                                    ymax=10, ctc_cutoff_Ang=4,
                                                                    representatives={"n_frames": 3, "scheme": "mean"})
        #fig.savefig("test.pdf")
        _plt.close("all")

    def test_repframes_dict_geoms(self):
        traj = _md.load(test_filenames.traj_xtc_stride_20, top=test_filenames.top_pdb)
        repframes = {"small" : traj[:3],
                     "big" : traj[:5]}
        fig, ax, sorted_keys, repframes_out = plots.compare_violins({"small": self.CGL394, "big": self.CGL394_larger},
                                                                    anchor="L394",
                                                                    ymax=10, ctc_cutoff_Ang=4,
                                                                    representatives=repframes)
        for key, val in repframes_out.items():
            assert val is repframes[key]
        #fig.savefig("test.pdf")
        _plt.close("all")


from mdciao.flare import _utils
import numpy as np
from unittest import TestCase
from mdciao import flare
from itertools import combinations
from unittest import skip
from mdciao.plots.plots import _colorstring
from mdciao.examples import filenames as test_filenames
import mdtraj as md
from matplotlib import pyplot as plt



class TestFlare(TestCase):

    def test_works(self):
        myax, _, _ = flare.freqs2flare([1, 1, 1],
                                    np.array([[0, 1], [1, 2], [2, 3]]),
                                    exclude_neighbors=0,
                                    )

        myax.figure.tight_layout()
        #myax.figure.savefig("test.png")

    def test_options_only_curves(self):
        myfig = plt.figure(figsize=(5,5))
        iax = plt.gca()
        iax , _, _ = flare.freqs2flare(np.array([[1, 1, 1]]),
                                       np.array([[0, 1], [1, 2], [2, 3]]),
                                       fragments=[[0, 1], [2, 3]],
                                       exclude_neighbors=0,
                                       plot_curves_only=True,
                                       ax=iax,
                                       )

        iax.figure.tight_layout()
        #ax.figure.savefig("test.png")

    def test_options_sparse(self):
        myfig, myax = plt.subplots(1, 2, sharex=True, sharey=True, figsize=(10, 5))
        iax, _, _ = flare.freqs2flare(np.array([[1, 1, 1]]),
                                      np.array([[0, 1], [1, 2], [2, 3]]),
                                      fragments=[[0, 1], [2, 3], np.arange(4, 15)],
                                      exclude_neighbors=0,
                                      ax=myax[0]

                                      )
        iax, _, _ = flare.freqs2flare(np.array([[1, 1, 1]]),
                                      np.array([[0, 1], [1, 2], [2, 3]]),
                                      fragments=[[0, 1], [2, 3], np.arange(4, 15)],
                                      exclude_neighbors=0,
                                      ax=myax[1],
                                      sparse_residues=True,
                                      subplot=True,

                                      )

        iax.figure.tight_layout()
        #ax.figure.savefig("test.png")

    def test_coarse_grain(self):
        myfig = plt.figure(figsize=(5,5))
        iax = plt.gca()
        iax , _, _ = flare.freqs2flare(np.array([[1, 1, 1]]),
                                       np.array([[0, 1], [1, 2], [2, 3]]),
                                       fragments=[[0, 1], [2, 3]],
                                       coarse_grain=True,
                                       ax=iax,
                                       )
class TestCirclePlotResidues(TestCase):

    def test_works(self):
        iax, _, cpr_dict =  flare.circle_plot_residues([np.arange(50),
                                                        np.arange(50, 100)],
                                                       ss_array=["H"] * 100,
                                                       fragment_names=["A", "B"],
                                                       textlabels={0:"first",99:"last"},
                                                       aura=np.arange(100),
                                                       )
        for key in ["fragment_labels",
                    "dot_labels",
                    "dots",
                    "SS_labels"
                    ]:
            assert key in cpr_dict.keys()

        #ax.figure.savefig("test.png")

class TestAddBezier(TestCase):

    def test_works(self):
        iax, xy, cpr_dict = flare.circle_plot_residues([np.arange(5),
                                             np.arange(5, 10)],
                                            #ss_array=["H"] * 10,
                                            #fragment_names=["A", "B"]
                                             )
        pairs = [[0,5],
                 [5,6],
                 [9,3]
                 ]
        node_pairs = [(xy[ii],xy[jj]) for (ii,jj) in pairs]
        bzcurves =  flare.add_bezier_curves(iax,node_pairs
                                          )
        plt.close("all")
        #ax.figure.savefig("test.png")

class TestChord(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.pairs = np.reshape(np.arange(40), (2, -1)).T
        print(cls.pairs)
        cls.frags = np.reshape(np.arange(49), (7, 7))
        # pairs go up to 40, frags up to 49, that frag will disappear
        print(cls.frags)
        cls.freqs = np.arange(20)
        cls.fragment_names = ["A", "B", "C", "D", "E", "F", "G"]
        print(cls.freqs)
        print(cls.freqs.sum())

    def test_just_works(self):

        iax, nonzeros, plotattrs = flare.freqs2chord(self.freqs,
                                                     self.pairs,
                                                     self.frags,
                                                     fragment_names=self.fragment_names)
        assert isinstance(iax, plt.Axes)
        assert len(iax.texts) == 6 + 6
        print(plotattrs["sigmas"])
        np.testing.assert_array_equal(nonzeros, [0, 1, 2, 3, 4, 5])
        for key in ["fragment_names", "fragment_labels", "r", "dots", "sigmas"]:
            assert key in plotattrs.keys()
        np.testing.assert_array_equal(plotattrs["sigmas"], [21, 70, 99, 28, 77, 85])
        assert plotattrs["dots"][0].radius > 0
        #ax.figure.savefig("test.ref.pdf")
        plt.close("all")

    def test_one_less_fragment(self):

        iax, nonzeros, plotattrs = flare.freqs2chord(self.freqs,
                                                     self.pairs,
                                                     self.frags,
                                                     fragment_names=self.fragment_names,
                                                     min_sigma=25,
                                                     )
        """
        This is how the coarse grain matrix looks like, 
        row nr. 0 wil be dropped
        [[ 0.  0.  0. 21.  0.  0.  0.]
         [ 0.  0.  0.  7. 63.  0.  0.]
         [ 0.  0.  0.  0. 14. 85.  0.]
         [21.  7.  0.  0.  0.  0.  0.]
         [ 0. 63. 14.  0.  0.  0.  0.]
         [ 0.  0. 85.  0.  0.  0.  0.]
         [ 0.  0.  0.  0.  0.  0.  0.]]

        """
        #ax.figure.savefig("test.one_less.pdf")
        print(plotattrs["sigmas"])
        assert isinstance(iax, plt.Axes)
        assert len(iax.texts) == 5 + 5
        np.testing.assert_array_equal(nonzeros, [1, 2, 3, 4, 5])
        for key in ["fragment_names", "fragment_labels", "r", "dots", "sigmas"]:
            assert key in plotattrs.keys()
        np.testing.assert_array_equal(plotattrs["sigmas"], [# A isn't there bc 21 < 25
                                                            70, 99,
                                                            7, # note that frag-D has lost 21 cts from A
                                                            77, 85])
        assert plotattrs["dots"][0].radius > 0
        plt.close("all")


    def test_half_circle(self):

        iax, nonzeros, plotattrs = flare.freqs2chord(self.freqs,
                                                     self.pairs,
                                                     self.frags,
                                                     fragment_names=self.fragment_names,
                                                     normalize_to_sigma=self.freqs.sum()*2,
                                                     )
        #ax.figure.savefig("test.half.pdf")
        assert isinstance(iax, plt.Axes)
        assert len(iax.texts) == 6 + 6
        np.testing.assert_array_equal(nonzeros, [0, 1, 2, 3, 4, 5])
        for key in ["fragment_names", "fragment_labels", "r", "dots", "sigmas"]:
            assert key in plotattrs.keys()
        np.testing.assert_array_equal(plotattrs["sigmas"], [21, 70, 99, 28, 77, 85])
        assert plotattrs["dots"][0].radius > 0
        plt.close("all")
        print(plotattrs["sigmas"])

    def test_counter_clockwise(self):

        iax, nonzeros, plotattrs = flare.freqs2chord(self.freqs,
                                                     self.pairs,
                                                     self.frags,
                                                     fragment_names=self.fragment_names,
                                                     clockwise=False,
                                                     )
        #iax.figure.savefig("test.ccwise.pdf")
        assert isinstance(iax, plt.Axes)
        assert len(iax.texts) == 6 + 6
        np.testing.assert_array_equal(nonzeros, [0, 1, 2, 3, 4, 5])
        for key in ["fragment_names", "fragment_labels", "r", "dots"]:
            assert key in plotattrs.keys()
        np.testing.assert_array_equal(plotattrs["sigmas"], [21, 70, 99, 28, 77, 85])
        assert plotattrs["dots"][0].radius > 0
        plt.close("all")
        print(plotattrs["sigmas"])


from mdciao.flare import _utils
import numpy as np
from unittest import TestCase
from itertools import combinations
from unittest import skip
from mdciao.plots.plots import _colorstring
from mdciao.examples import filenames as test_filenames
import mdtraj as md
from matplotlib import pyplot as plt, lines as mpllines
from mdciao.flare import circle_plot_residues


class TestAngulateSegments(TestCase):

    def test_works(self):
        angles = _utils.regspace_angles(10, circle=360)

        np.testing.assert_array_equal(angles,
                                      [0., -36., -72., -108., -144., -180., -216., -252., -288., -324.])

    def test_anti_clockwise(self):
        angles = _utils.regspace_angles(10, circle=360, clockwise=False)

        np.testing.assert_array_equal(angles,
                                      [0, 36, 72, 108, 144, 180, 216, 252, 288, 324])


class TestFragmentSelectionParser(TestCase):

    def setUp(self):
        self.fragments = [[0, 1], [2, 3], [4, 5]]
        self.all_pairs = np.vstack(list(combinations(np.arange(6), 2)))

    def test_works(self):
        condition = _utils.fragment_selection_parser(self.fragments)
        assert all([condition(pair) for pair in self.all_pairs])

    def test_show_only(self):
        condition = _utils.fragment_selection_parser(self.fragments, only_show_fragments=[2])
        shown_pairs = [pair for pair in self.all_pairs if condition(pair)]
        np.testing.assert_array_equal([[0, 4],
                                       [0, 5],
                                       [1, 4],
                                       [1, 5],
                                       [2, 4],
                                       [2, 5],
                                       [3, 4],
                                       [3, 5],
                                       [4, 5]],
                                      shown_pairs)

    def test_hide(self):
        condition = _utils.fragment_selection_parser(self.fragments, hide_fragments=[0])
        shown_pairs = [pair for pair in self.all_pairs if condition(pair)]
        np.testing.assert_array_equal([[2, 3],
                                       [2, 4],
                                       [2, 5],
                                       [3, 4],
                                       [3, 5],
                                       [4, 5]],
                                      shown_pairs)

    def test_raises(self):
        with np.testing.assert_raises(ValueError):
            _utils.fragment_selection_parser(self.fragments,
                                             hide_fragments=[0], only_show_fragments=[1])


class TestCartify(TestCase):

    def test_works(self):
        XY, angles = _utils.cartify_fragments([[0, 1, 2, 4]],
                                              return_angles=True)

        np.testing.assert_array_almost_equal(angles / np.pi * 180, [0, -90, -180, -270])
        np.testing.assert_array_almost_equal(XY, [[1, 0],
                                                  [0, -1],
                                                  [-1, 0],
                                                  [0, 1],
                                                  ])

    def test_padding_between(self):
        XY = _utils.cartify_fragments([[0], [1]],
                                      padding=[0, 1, 0],  # The first and 3rd positions should be empty
                                      )
        np.testing.assert_array_almost_equal(XY,
                                             [
                                                 [0, -1],
                                                 [0, 1],
                                             ])

    def test_padding_initial(self):
        XY = _utils.cartify_fragments([[0, 1]],
                                      padding=[2, 0, 0],
                                      )
        np.testing.assert_array_almost_equal(XY, [[-1, 0],
                                                  [0, 1],
                                                  ])

    def test_padding_final(self):
        XY = _utils.cartify_fragments([[0, 1]],
                                      padding=[0, 0, 2],
                                      )
        np.testing.assert_array_almost_equal(XY, [[1, 0],
                                                  [0, -1],
                                                  ])


class TestColors(TestCase):

    def setUp(self):
        self.ref_colors = _colorstring.split(",")
        self.fragments = [[0, 1], [2, 3], [4, 5]]

    def test_works_False(self):
        colorlist = _utils.col_list_from_input_and_fragments(False, self.fragments)
        np.testing.assert_array_equal(["gray"] * 6, colorlist)

    def test_works_True(self):
        colorlist = _utils.col_list_from_input_and_fragments(True, self.fragments)
        np.testing.assert_array_equal([self.ref_colors[0], self.ref_colors[0],
                                       self.ref_colors[1], self.ref_colors[1],
                                       self.ref_colors[2], self.ref_colors[2]],
                                      colorlist)

    def test_works_string(self):
        colorlist = _utils.col_list_from_input_and_fragments("salmon", self.fragments)
        np.testing.assert_array_equal(['salmon'] * 6,
                                      colorlist)

    def test_works_iterable_len_fragments(self):
        colorlist = _utils.col_list_from_input_and_fragments(["r", "g", "b"], self.fragments)
        np.testing.assert_array_equal(["r", "r",
                                       "g", "g",
                                       "b", "b"],
                                      colorlist)

    def test_works_iterable_len_residues(self):
        colorlist = _utils.col_list_from_input_and_fragments(["r", "r", "g", "g", "b", "b"], self.fragments)
        np.testing.assert_array_equal(["r", "r", "g", "g", "b", "b"],
                                      colorlist)

    def test_raises_wrong_length(self):
        with self.assertRaises(ValueError):
            _utils.col_list_from_input_and_fragments(["r", "r", "g", "g", "b"], self.fragments)

    def test_raises_wrong_type(self):
        with self.assertRaises(Exception):
            _utils.col_list_from_input_and_fragments(1, self.fragments)


    def test_works_dict(self):
        colorlist = _utils.col_list_from_input_and_fragments({"frag1": "r",
                                                              "frag2": "g",
                                                              "frag3": "b"},
                                                             self.fragments)
        np.testing.assert_array_equal(["r", "r",
                                       "g", "g",
                                       "b", "b"],
                                      colorlist)

    def test_works_one_frag(self):
        colorlist = _utils.col_list_from_input_and_fragments(False, [0, 1, 2])
        np.testing.assert_array_equal(["gray"] * 3, colorlist)


class TestLambaCurve(TestCase):

    def setUp(self):
        self.fragments = [[0, 1],
                          [2, 3],
                          [4, 5],
                          [6, 7]]
        self.pairs = np.vstack(list(combinations(range(8), 2)))

    def test_works_all_pairs(self):
        ilambda = _utils.should_this_residue_pair_get_a_curve(self.fragments)
        assert all([ilambda(pair) for pair in self.pairs])

    def test_res_selection(self):
        res_selection = [1, 4]
        ilambda = _utils.should_this_residue_pair_get_a_curve(self.fragments,
                                                              select_residxs=res_selection)
        selected_pairs = np.vstack([pair for pair in self.pairs if ilambda(pair)])
        np.testing.assert_array_equal(selected_pairs, [[0, 1],  # bc 1
                                                       [0, 4],  # bc 4
                                                       [1, 2],  # bc 1
                                                       [1, 3],  #
                                                       [1, 4],
                                                       [1, 5],
                                                       [1, 6],
                                                       [1, 7],  # bc 1
                                                       [2, 4],  # bc 4
                                                       [3, 4],
                                                       [4, 5],
                                                       [4, 6],
                                                       [4, 7]
                                                       ])

    def test_frag_selection(self):
        # This is tested anyway in its own lambda but testing again here
        ilambda = _utils.should_this_residue_pair_get_a_curve(self.fragments,
                                                              mute_fragments=[0, 1])
        selected_pairs = np.vstack([pair for pair in self.pairs if ilambda(pair)])
        np.testing.assert_array_equal(selected_pairs, [[4, 5],
                                                       [4, 6],
                                                       [4, 7],
                                                       [5, 6],
                                                       [5, 7],
                                                       [6, 7]
                                                       ])

    def test_neighbor_selection(self):
        ilambda = _utils.should_this_residue_pair_get_a_curve(self.fragments,
                                                              exclude_neighbors=3)
        selected_pairs = np.vstack([pair for pair in self.pairs if ilambda(pair)])
        np.testing.assert_array_equal(selected_pairs, [[0, 4],
                                                       [0, 5],
                                                       [0, 6],
                                                       [0, 7],
                                                       [1, 5],
                                                       [1, 6],
                                                       [1, 7],
                                                       [2, 6],
                                                       [2, 7],
                                                       [3, 7],
                                                       ]
                                      )

    def test_neighbor_selection_w_top(self):
        top = md.load(test_filenames.actor_pdb).top # All is one chain here
        ilambda = _utils.should_this_residue_pair_get_a_curve(self.fragments,
                                                              top=top,
                                                              exclude_neighbors=3)
        selected_pairs = np.vstack([pair for pair in self.pairs if ilambda(pair)])
        np.testing.assert_array_equal(selected_pairs, [[0, 4],
                                                       [0, 5],
                                                       [0, 6],
                                                       [0, 7],
                                                       [1, 5],
                                                       [1, 6],
                                                       [1, 7],
                                                       [2, 6],
                                                       [2, 7],
                                                       [3, 7],
                                                       ]
                                      )


class TestResidueLabels(TestCase):

    def test_works(self):
        plt.figure(figsize=(5, 5))
        iax = plt.gca()
        iax.set_aspect("equal")
        iax.set_xlim([-1, 1])
        iax.set_ylim([-1, 1])
        _utils.add_residue_labels(plt.gca(), [[1, 0],
                                              [0, -1],
                                              [-1, 0],
                                              [0, 1]],
                                  [15, 18, 21, 12],
                                  10,
                                  center=[0, 0])
        ifig = plt.gcf()
        ifig.tight_layout()
        # plt.savefig('test.png',bbox_inches="tight")

    def test_works_options(self):
        top = md.load(test_filenames.actor_pdb).top
        ifig, myax = plt.subplots(1, 2, sharex=True, sharey=True, figsize=(10, 5))
        iax = plt.gca()
        iax.set_aspect("equal")
        iax.set_xlim([-1, 1])
        iax.set_ylim([-1, 1])
        _utils.add_residue_labels(myax[0], [[1, 0],
                                            [0, -1],
                                            [-1, 0],
                                            [0, 1]],
                                  [15, 18, 21, 12],
                                  10,
                                  center=[0, 0],
                                  top=top,
                                  colors=['b'] * 4)

        _utils.add_residue_labels(myax[1], [[1, 0],
                                            [0, -1],
                                            [-1, 0],
                                            [0, 1]],
                                  [15, 18, 21, 12],
                                  10,
                                  center=[0, 0],
                                  top=top,
                                  highlight_residxs=[15],
                                  replacement_labels={18: "6 o'clock"},
                                  shortenAAs=False)

        ifig.tight_layout()
        # plt.savefig('test.png',bbox_inches="tight")


class TestFragmentLabels(TestCase):

    def test_works(self):
        plt.figure(figsize=(5, 5))
        iax = plt.gca()
        iax.set_aspect("equal")
        iax.set_xlim([-1, 1])
        iax.set_ylim([-1, 1])
        fragments = [np.arange(10), np.arange(10, 20)]
        _, _, plattrb = circle_plot_residues(fragments,
                                             ax=iax,
                                             padding=[0, 0, 0])

        _utils.add_fragment_labels(fragments, ["frag_10_20", "frag_30_40"],
                                   iax,
                                   r=plattrb["r"])
        ifig = plt.gcf()
        ifig.tight_layout()
        # plt.savefig('test.png',bbox_inches="tight")
        plt.close("all")

    def _test_works_options(self):
        top = md.load(test_filenames.actor_pdb).top
        ifig, myax = plt.subplots(1, 2, sharex=True, sharey=True, figsize=(10, 5))
        iax = plt.gca()
        iax.set_aspect("equal")
        iax.set_xlim([-1, 1])
        iax.set_ylim([-1, 1])
        _utils.add_residue_labels(myax[0], [[1, 0],
                                            [0, -1],
                                            [-1, 0],
                                            [0, 1]],
                                  [0,
                                   -np.pi / 2,
                                   -np.pi,
                                   -np.pi * 1.5],
                                  [15, 18, 21, 12],
                                  10,
                                  top=top,
                                  colors=['b'] * 4)

        _utils.add_residue_labels(myax[1], [[1, 0],
                                            [0, -1],
                                            [-1, 0],
                                            [0, 1]],
                                  [0,
                                   -np.pi / 2,
                                   -np.pi,
                                   -np.pi * 1.5],
                                  [15, 18, 21, 12],
                                  10,
                                  top=top,
                                  highlight_residxs=[15],
                                  replacement_labels={18: "6 o'clock"},
                                  shortenAAs=False)

        ifig.tight_layout()
        # plt.savefig('test.png',bbox_inches="tight")


class Testvalue2pos(TestCase):
    def test_works(self):
        x = [10, 20, 3, 5, 7, 1000, 0]

        v2p = _utils.value2position_map(x)
        assert 0 == v2p[10]
        assert 1 == v2p[20]
        assert 2 == v2p[3]
        assert 3 == v2p[5]
        assert 4 == v2p[7]
        assert 5 == v2p[1000]
        assert 6 == v2p[0]


class TestMyBezier(TestCase):

    def test_create(self):
        mybz = _utils.create_flare_bezier_2([[0, -1], [1, 0]], [0, 0])
        assert isinstance(mybz, _utils.my_BZCURVE)

    def test_plot(self):
        mybz = _utils.create_flare_bezier_2(np.array([[0, -1], [1, 0]]), [0, 0])
        mybz.plot(50)
        assert isinstance(mybz.Line2D, mpllines.Line2D)


class Test_parse_residue_and_fragments(TestCase):

    def setUp(self):
        self.res_idx_pairs = np.array([[1, 2],
                                       [1, 3],
                                       [1, 4],
                                       [2, 4],
                                       [10, 11],
                                       [10, 12],
                                       [10, 15]])
        self.top = md.load(test_filenames.actor_pdb).top

    def test_works(self):
        res_idxs_as_fragments, anchors, muted = _utils._parse_residue_and_fragments(self.res_idx_pairs)
        np.testing.assert_equal(len(res_idxs_as_fragments), 1)
        np.testing.assert_array_equal(res_idxs_as_fragments, [np.arange(0, 16)])
        assert anchors is None
        assert muted is None

    def test_sparse_residues_True(self):
        res_idxs_as_fragments, _, _  = _utils._parse_residue_and_fragments(self.res_idx_pairs, sparse_residues=True)
        np.testing.assert_equal(len(res_idxs_as_fragments), 1)
        np.testing.assert_array_equal(res_idxs_as_fragments, [[1, 2, 3, 4, 10, 11, 12, 15]])

    def test_sparse_residues_value(self):
        res_idxs_as_fragments, _, _  = _utils._parse_residue_and_fragments(self.res_idx_pairs, sparse_residues=[1, 10, 20])
        np.testing.assert_equal(len(res_idxs_as_fragments), 1)
        np.testing.assert_array_equal(res_idxs_as_fragments, [[1, 10, 20]])

    def test_sparse_residues_top(self):
        res_idxs_as_fragments, _, _ = _utils._parse_residue_and_fragments(self.res_idx_pairs, top=self.top)
        np.testing.assert_equal(len(res_idxs_as_fragments), 1)
        np.testing.assert_array_equal(res_idxs_as_fragments, [np.arange(self.top.n_residues)])

    def test_fragments(self):
        fragments = [[0],
                     [1, 2, 3, 4, 5],
                     [6, 7, 8, 9],
                     [10, 11, 12, 13, 14, 15],
                     [20, 21, 50]]
        res_idxs_as_fragments, _, _  = _utils._parse_residue_and_fragments(self.res_idx_pairs, fragments=fragments)
        np.testing.assert_equal(len(res_idxs_as_fragments), len(fragments))
        self.assertListEqual(res_idxs_as_fragments, fragments)

    def test_fragments_sparse_residues(self):
        fragments = [[0],
                     [1, 2, 3, 4, 5],
                     [6, 7, 8, 9],
                     [10, 11, 12, 13, 14, 15],
                     [20, 21, 50]]
        res_idxs_as_fragments, _, _  = _utils._parse_residue_and_fragments(self.res_idx_pairs, sparse_residues=True,
                                                                    fragments=fragments)
        np.testing.assert_equal(len(res_idxs_as_fragments), 2)
        self.assertListEqual(res_idxs_as_fragments, [[1, 2, 3, 4], [10, 11, 12, 15]])

    def test_fragments_sparse_residues_w_anchor_and_mute(self):
        fragments = [[0],
                     [1, 2, 3, 4, 5],
                     [6, 7, 8, 9],
                     [10, 11, 12, 13, 14, 15],
                     [20, 21, 50]]
        res_idxs_as_fragments, new_anchors, new_mutes  = _utils._parse_residue_and_fragments(self.res_idx_pairs,
                                                                           sparse_residues=True,
                                                                           fragments=fragments,
                                                                           anchor_fragments=[1],
                                                                           mute_fragments=[3])
        np.testing.assert_equal(len(res_idxs_as_fragments), 2)
        np.testing.assert_array_equal(res_idxs_as_fragments, [[1, 2, 3, 4], [10, 11, 12, 15]])
        np.testing.assert_equal(new_anchors,[0]) #old was 1
        np.testing.assert_equal(new_mutes,[1])# old was 3

    def test_fragments_sparse_residues_w_anchor_and_mute_None(self):
        fragments = [[0],
                     [1, 2, 3, 4, 5],
                     [6, 7, 8, 9],
                     [10, 11, 12, 13, 14, 15],
                     [20, 21, 50]]
        res_idxs_as_fragments, new_anchors, new_mutes  = _utils._parse_residue_and_fragments(self.res_idx_pairs,
                                                                           sparse_residues=True,
                                                                           fragments=fragments,
                                                                           anchor_fragments=[0],
                                                                           mute_fragments=[3])
        np.testing.assert_equal(len(res_idxs_as_fragments), 2)
        np.testing.assert_array_equal(res_idxs_as_fragments, [[1, 2, 3, 4], [10, 11, 12, 15]])
        np.testing.assert_equal(new_anchors,None) #old was 0, frag 0 got deleted
        np.testing.assert_equal(new_mutes,[1])# old was 3

    def test_fragments_sparse_residues_value(self):
        fragments = [[0],
                     [1, 2, 3, 4, 5],
                     [6, 7, 8, 9],
                     [10, 11, 12, 13, 14, 15],
                     [20, 21, 50]]
        res_idxs_as_fragments = _utils._parse_residue_and_fragments(self.res_idx_pairs,
                                                                    sparse_residues=[1, 10, 20],
                                                                    fragments=fragments)[0]
        np.testing.assert_equal(len(res_idxs_as_fragments), 3)
        np.testing.assert_array_equal(res_idxs_as_fragments, [[1], [10], [20]])

    def test_fragments_sparse_fragments(self):
        fragments = [[0],
                     [1, 2, 3, 4, 5],
                     [6, 7, 8, 9],
                     [10, 11, 12, 13, 14, 15],
                     [20, 21, 50]]
        self.res_idx_pairs = np.array([[1, 2],
                                       [1, 3],
                                       [1, 4],
                                       [2, 4],
                                       [10, 11],
                                       [10, 12],
                                       [10, 15]])
        res_idxs_as_fragments = _utils._parse_residue_and_fragments(self.res_idx_pairs,
                                                                    fragments=fragments,
                                                                    sparse_fragments=True)[0]
        np.testing.assert_equal(len(res_idxs_as_fragments), 2)
        self.assertListEqual(res_idxs_as_fragments, [[1, 2, 3, 4, 5],
                                                     [10, 11, 12, 13, 14, 15]]
                             )

class Test_Aura(TestCase):

    def test_works(self):
        iax, xy, cpr_dict = circle_plot_residues([np.arange(50),
                                                  np.arange(50, 100)],
                                                 )
        _utils.add_aura(xy, np.mod(np.arange(len(xy)), 3), iax, r=cpr_dict["r"])
        # ax.figure.savefig("test.png")
        plt.close("all")

    def test_works_w_options(self):
        iax, xy, cpr_dict = circle_plot_residues([np.arange(50),
                                                  np.arange(50, 100)],
                                                 )
        _utils.add_aura(xy, np.mod(np.arange(len(xy)), 3), iax, r=cpr_dict["r"],
                        lines=False)
        # ax.figure.savefig("test.png")
        plt.close("all")
    def test_negative(self):
        iax, xy, cpr_dict = circle_plot_residues([np.arange(50),
                                                  np.arange(50, 100)],
                                                 )
        neg = np.ones(len(xy))
        neg[int(len(neg)/2):]=-1
        _utils.add_aura(xy, neg*np.mod(np.arange(len(xy)), 3), iax, r=cpr_dict["r"],
                        lines=True)
        #ax.figure.savefig("test.png")
        plt.close("all")


class TestGetSetFonts(TestCase):

    def test_get(self):
        iax, _, _ = circle_plot_residues([np.arange(20)])

        fs = _utils.fontsize_get(iax)
        self.assertSequenceEqual(["n_polygons", "other"], list(fs.keys()))
        assert len(fs["other"]) == 0
        assert len(fs["n_polygons"]) == 1
        print(np.unique([tt.get_fontsize() for tt in iax.texts]))
        self.assertSequenceEqual(fs["n_polygons"], np.unique([np.round(tt.get_fontsize(), 2) for tt in iax.texts]))

    def test_set(self):
        iax1, _, _ = circle_plot_residues([np.arange(20)])
        iax2, _, _ = circle_plot_residues([np.arange(40)])

        fs1, fs2 = [_utils.fontsize_get(iax)["n_polygons"][0] for iax in [iax1, iax2]]
        assert fs1 != fs2

        _utils.fontsize_apply(iax1, iax2)
        new_fs2 = _utils.fontsize_get(iax2)["n_polygons"][0]
        assert fs1 == new_fs2

class Test_coarse_grain_freqs_by_frag(TestCase):

    def setUp(self):
        pass

    def test_works(self):
        freqs = [1, 2,
                 4, 5,
                 6]

        frags = [[0, 1], [2, 3], [4, 5]]

        pairs = [[0, 2], [1, 3],  # frags 0-1
                 [0, 4], [1, 5],  # frags 0-2
                 [2, 4]]  # frags 1-2

        ref_mat = np.zeros((3,3))
        ref_mat[0,1]=ref_mat[1,0]=3
        ref_mat[0,2]=ref_mat[2,0]=9
        ref_mat[1,2]=ref_mat[2,1]=6

        mat = _utils.coarse_grain_freqs_by_frag(freqs, pairs, frags)
        np.testing.assert_array_equal(mat, ref_mat)

    def test_raises_orphan(self):
        freqs = [1, 2,
                 4, 5,
                 6]

        frags = [[0, 1], [2, 3], [4, 5]]

        pairs = [[0, 2], [1, 3],  # frags 0-1
                 [0, 4], [1, 6],  # frags 0-2 and 0-None
                 [2, 4]]  # frags 1-2

        with np.testing.assert_raises(ValueError):
            _utils.coarse_grain_freqs_by_frag(freqs, pairs, frags)

    def test_ignores_orphan(self):
        freqs = [1, 2,
                 4, 5,
                 6]

        frags = [[0, 1], [2, 3], [4, 5]]

        pairs = [[0, 2], [1, 3],  # frags 0-1
                 [0, 4], [1, 6],  # frags 0-2 and 0-None
                 [2, 4]]  # frags 1-2

        ref_mat = np.zeros((3,3))
        ref_mat[0,1]=ref_mat[1,0]=3
        ref_mat[0,2]=ref_mat[2,0]=4 #since the second pair didn't get summed
        ref_mat[1,2]=ref_mat[2,1]=6
        mat = _utils.coarse_grain_freqs_by_frag(freqs, pairs, frags, check_if_subset=False)
        np.testing.assert_array_equal(mat, ref_mat)

class Test_sparsify_sym_matrix(TestCase):

    def test_just_works(self):
        mat = np.array([[1, 0, 2],
                        [0, 0, 0],
                        [2, 0, 3]])
        nm, nz = _utils.sparsify_sym_matrix(mat)
        np.testing.assert_array_equal(nz, [0,2])
        np.testing.assert_array_equal(nm, [[1,2],
                                           [2,3]])

class Test_sparsify_sym_matrix_by_row_sum(TestCase):

    def test_just_works(self):
        #First row.sum() is below eps, will be eliminated
        # Because matrix is sym, first column also goes
        mat = np.array([[0, 0, 0.01],
                        [0, .02, .02],
                        [0.01, .02, 0.0]])
        sparse_mat, kept_idxs, lost_value = _utils.sparsify_sym_matrix_by_row_sum(np.array(mat), eps=1e-2)
        np.testing.assert_array_equal([[0.02, 0.02],
                                       [0.02, 0.0]], sparse_mat)
        np.testing.assert_array_equal([1,2], kept_idxs)
        np.testing.assert_almost_equal(0.01, lost_value)


##############################################################################
#    This file is part of mdciao.
#
#    Copyright 2024 Charité Universitätsmedizin Berlin and the Authors
#
#    Authors: Guillermo Pérez-Hernandez
#    Contributors:
#
#    mdciao is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    mdciao is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with mdciao.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

import numpy as _np
from time import time as _time
import threading as _threading
from multiprocessing import Manager as _Manager
from IPython.display import display as _display
import ipywidgets as _ipywidgets
from time import sleep as _sleep
from datetime import timedelta as _timedelta
import signal as _signal
from shutil import get_terminal_size as _get_terminal_size

# My attempt at a useful progress reporting for parallel processes.
# The main method is _prepare_progressbar_thread, which generates all objects
# needed to handle asyncronous progress report to the standard output (for the
# terminal or to textwidget in the notebook

def _is_notebook() -> bool:
    # https://stackoverflow.com/a/39662359 and also the answer below (like autonotebook.py)
    try:
        from IPython import get_ipython as _get_ipython #lazy import here is better, can except later
        shell = _get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True  # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except (NameError, ModuleNotFoundError):
        return False  # Probably standard Python interpreter

def _progress_dict2infoline(idict, first_update_after=1.5) -> str:
    r"""
    Format the progress-report string from a progress dictionary.

    The progress dictionary is a managed dictionary created by the
    method _prepare_progressbar_thread. It gets passed to the
    individual per-trajectory processes that update its fields as
    they stream through trajectories in parallel (per_traj_ctc and
    per_traj_mindist_lower_bound)

    Meanwhile, a separate thread, which also has access to the
    progress dictionary 1) updates the progress report line using
    this method and takes care of printing w/ or w/o overwriting
    lines in the standard output.

    Both the individual per-trajectory processes (per_traj_ctc and
    per_traj_mindist_lower_bound) as well as the methods that
    called them(trajs2ctcs and  trajs2lower_bounds, respectively)
    used this method to compuse the `infoline`.

    The parameters computed in this method for the `infoline`
    are:
    * n_trajs_done/n_trajs_total
    * Elapsed time
    * remaining time
    * Trajs/s
    * Frames/s

    The `infoline` looks like this:

    `Processing trajectories: 1/4 [ 25%]. Frames processed:    32001. Elapsed time:  0:00:41. Remaining time ~  0:02:04. Trajs/s:    0. Frames/s:  777`

    Please note, the Trajs/s ad Frames/s values are NOT
    instantaneous but overall computation, taking IO
    time into account (measured time is walltime, not
    process time).


    Parameters
    ----------
    idict : dict
        A managed dictionary containing  values needed to
        compile performance parameters to be printed
        in plaintext as a progressbar analog. The dictionary
        needs to have the fields:
        * 'start_time'
        * 'n_trajs_total'
        * 'n_trajs_done'
        * 'n_frames_done'

    first_update_after : float
        In seconds, how much to wait before computing
        performance values (Frames/sec

    Returns
    -------
    infoline : str
        A string with information about the performance
        of the underlying computations. It looks like this:
        `Processing trajectories: 1/4 [ 25%]. Frames processed:    32001. Elapsed time:  0:00:41. Remaining time ~  0:02:04. Trajs/s:    0. Frames/s:  777`
    """
    elapsed = _time() - idict['start_time']
    n = len(str(idict['n_trajs_total']))
    perc = int(_np.round(idict['n_trajs_done'] / idict['n_trajs_total'] * 100))

    if elapsed < first_update_after and idict['n_trajs_done']!=idict["n_trajs_total"]:
        elapsed = 'hh:mm:ss'
        remaining = 'hh:mm:ss'
        trajs_per_s = ''
    else:
        try:
            # Only recompute frames_per_s if more frames have been processed
            # Otherwise if n_frames_done stays the same but the clock is counting,
            # then frames_per_sec necesarily decreases
            if idict["n_frames_done"]>idict["n_frames_done_prev"]:
                idict["frames_per_s"] = int(_np.round(idict['n_frames_done'] / elapsed))
                idict["n_frames_done_prev"]=idict["n_frames_done"]
        except TypeError:
            idict["frames_per_s"] = ''
            idict["n_frames_done"] = ''
        trajs_per_s = int(_np.round(idict['n_trajs_done'] / elapsed))
        try:
            avg_t_traj = elapsed / idict['n_trajs_done']
            remaining = _timedelta(seconds=_np.round(avg_t_traj * (idict['n_trajs_total'] - idict['n_trajs_done'])))
        except ZeroDivisionError:
            remaining = 'hh:mm:ss'
        elapsed = _timedelta(seconds=_np.round(elapsed))
    return f"Processing trajectories: {idict['n_trajs_done'] :{n}}/{idict['n_trajs_total']} [{perc :3}%]. " \
           f"Frames processed: {idict['n_frames_done'] :8}. " \
           f"Elapsed time: {str(elapsed) :>8}. " \
           f"Remaining time ~ {str(remaining) :>8}. " \
           f"Trajs/s: {trajs_per_s  :4}. " \
           f"Frames/s: {idict['frames_per_s'] :4}."

def _progressbardict2thread(progressbar_dict, sleep_between_updates=1, overwrite=True):
    r"""
    Prepares threading objects (an Event and a Thread) as well as the target method of the thread

    Parameters
    ----------
    progressbar_dict : dict
        A managed, thread-safe dict
    sleep_between_updates : float, default is 1
        The frequency at which the progressbar updates
    overwrite : bool, default is True
        Whether to overwrite the output or not. Assumes the
        number of lines needed to be overwritten is
        the same as the lines to be writting.

    Returns
    -------
    thread : _threading.Thread
        A Thread-object with the correct target
        method which knows how to update the
        progress report lines.
    exit_event : _threading.Event
        An event that can be used
        as flag for when to stop updating
    """

    if _is_notebook():
        widg_len = max([len(bar) + 5 for bar in progressbar_dict["pbars"]])
        progress = _ipywidgets.Textarea(value="\n".join(progressbar_dict["pbars"]), rows=len(progressbar_dict["pbars"]),
                                        layout=_ipywidgets.Layout(width=f'{widg_len}ch'),
                                        style={'font-size': f'{16}em'}
                                        )

        # Somewhat from here https://github.com/jupyter-widgets/ipywidgets/issues/2206#issuecomment-483246874
        _display(_ipywidgets.HTML("<style> .no-border textarea { border: none; resize: none; min-width: %uch} </style>"%widg_len))
        progress.add_class("no-border")
        def work(progress):
            while not exit_event.is_set():
                _sleep(sleep_between_updates)
                progressbar_dict["pbars"][0] = _progress_dict2infoline(progressbar_dict)
                progress.layout.width = f'{max([len(bar) + 5 for bar in progressbar_dict["pbars"]])}ch'
                progress.value = "\n".join(progressbar_dict["pbars"])
            if exit_event.is_set():
                # print one last time
                progressbar_dict["pbars"][0] = _progress_dict2infoline(progressbar_dict)
                progress.layout.width = f'{max([len(bar) + 5 for bar in progressbar_dict["pbars"]])}ch'
                progress.value = "\n".join(progressbar_dict["pbars"])

        thread = _threading.Thread(target=work, args=(progress,))
        _display(progress)
    else:
        progress = None
        _print_w_option_to_overwrite(progressbar_dict["pbars"], overwrite=False)
        def work(progress):  # arg is just for compat

            while not exit_event.is_set():
                # Alternatively, one could move the print statements to the individual processes
                # launched by per_traj_ctc (passing a manager.Lock() to per_traj_ctc) s.t.
                # the print is called each time pbars is updated.
                # Instead, here we print at regular sleep(n) intervals
                _sleep(sleep_between_updates)
                progressbar_dict["pbars"][0] = _progress_dict2infoline(progressbar_dict)
                _print_w_option_to_overwrite(progressbar_dict["pbars"], overwrite=overwrite)
            if exit_event.is_set():
                # print one last time
                progressbar_dict["pbars"][0] = _progress_dict2infoline(progressbar_dict)
                _print_w_option_to_overwrite(progressbar_dict["pbars"], overwrite=overwrite)
        thread = _threading.Thread(target=work, args=(progress,))

    exit_event = _threading.Event()
    def handle_kb_interrupt(sig, frame):
        exit_event.set()
        thread.join()
        raise KeyboardInterrupt

    _signal.signal(_signal.SIGINT, handle_kb_interrupt)

    return thread, exit_event

def _print_w_option_to_overwrite(lines, overwrite=False):
    r"""

    Print `lines` with the option to overwrite the previous `len(lines)` lines before printing.

    Tries to take line-wrapping into account

    Todo use a decorator instead of if-else

    Parameters
    ----------
    lines : list
        The lines to be printed, joined
        with the newline "\n" character
    overwrite : bool
        Overwrite the last len(lines) lines,
        by moving the cursor up len(nlines) before
        printing
    """
    if overwrite:
        # Get the terminal size in columns/characters to find out the number of wrapped lines
        terminal_width = _get_terminal_size().columns
        n_wrapped_lines = _np.sum([len(line) > terminal_width for line in lines])
        # ANSI escape characters https://gist.github.com/fnky/458719343aabd01cfb17a3a4f7296797#escape
        for __ in range(len(lines) + n_wrapped_lines):
            print("\033[1A", end="\r", flush=True)
        print("\n".join(lines), flush=True)
    else:
        print("\n".join(lines), flush=True)

def _prepare_progressbar_thread(progressbar_dict, progressbar):
    r"""

    Prepare the objects needed for a progress report from independent processes
    to report their progress independently of each other.

    These objects are:
    * a managed dictionary, which is thread-safe and can be accessed
      by independent processes and/or threads.
      (https://docs.python.org/3/library/multiprocessing.html#multiprocessing.Manager)
      Apart from key,value pairs, the dictionary itself contains two managed lists:
       * `pbars`, a list strings for the independent progress report lines (one per process called by `delayed`)
       * `indices_of_free_pbars`, a list of booleans to check which `pbars` are free before using them to report
    * a threading.Thread-object, https://docs.python.org/3/library/threading.html#threading.Thread.
      It is instantiated with method called `work` as the `target` parameter (Thread(target=work)). The
      method `work` is created on-the-fly depending on whether we're updating to a terminal (update via print() to stdout)
      or to a notebook (update via a Textwidget).
    * a threading.Event-object, https://docs.python.org/3/library/threading.html#threading.Event.
      It is exit event; the flag that will tell the thread's target method (work) to exit the updating process. I.e.
      work contains a loop that runs with a sleep parameter and a while not exit_event

    The threading objects are created by _progressbardict2thread, which contains the updating logic
    and decisions in its environment dependent `work` method

    Parameters
    ----------
    progressbar_dict : dict
        A normal dictionary containing the fields needed to instantiate a managed
        progressbar dict. These fields are.
        * "n_trajs_total"
        * "n_trajs_done"
        * "n_trajs_done_prev"
        * "n_frames_done"
        * "start_time"
        * "n_jobs"
    progressbar : bool
        Toggle the progressbar. If False, all returned values
        will be None, telling downstream methods not to use
        a progressbar at all. This keeps the logic on whether
        to create a progressbar or not encapsulated here, and
        not in in the other methods.


    Returns
    -------
    progressbar_dict : _Manger().dict() or None
    thread : _threading.Thread or None
    exit_event : _threading.Event or None
    """
    if progressbar:
        # Managed variables (counters) for asynchronous progress bars
        manager = _Manager()
        progressbar_dict = manager.dict(progressbar_dict)
        pbars = manager.list([_progress_dict2infoline(progressbar_dict)] + ["" for ii in range(progressbar_dict["n_jobs"])])
        progressbar_dict.update({"pbars": pbars,
                                 "indices_of_free_pbars": manager.list([False] + [True] * progressbar_dict["n_jobs"])})

        thread, exit_event = _progressbardict2thread(progressbar_dict, sleep_between_updates=0.5)
        thread.start()
    else:
        print("Processing trajectories...", end="\r") #TODO putting this in the main method with decorator
        progressbar_dict, thread, exit_event = None, None, None,


    return progressbar_dict, thread, exit_event


import unittest

from mdciao import parsers

class TestParsersRun(unittest.TestCase):
    # This parsers will be tested anyway when testing the scripts,
    # but anyway, here we are
    def test_all(self):
        parsers.parser_for_frag_overview()
        parsers.parser_for_GPCR_overview()
        parsers.parser_for_densities()
        parsers.parser_for_CGN_overview()
        parsers.parser_for_rn()
        parsers.parser_for_interface()
        parsers.parser_for_dih()
        parsers.parser_for_sites()
        parsers.parser_for_compare_neighborhoods()
        parsers.parser_for_examples()
        parsers.parser_for_residues()

class Test_inform_of_parser(unittest.TestCase):

    def test_just_runs(self):
        p = parsers.parser_for_rn()
        parsers._inform_of_parser(p, [None,None])


if __name__ == '__main__':
    unittest.main()


import unittest
from mdciao.examples import filenames as test_filenames
from mdciao.sites import siteIO

import numpy as _np
import mdtraj as _md


import mdciao

class Test_x2site(unittest.TestCase):

    def test_runs(self):
        site = mdciao.sites.x2site(test_filenames.GDP_json)
        _np.testing.assert_equal(site["name"],"GDP")
        _np.testing.assert_equal(site["n_pairs"],5)
        _np.testing.assert_equal(len(site["pairs"]["AAresSeq"]),5)
        _np.testing.assert_array_equal(site["pairs"]["AAresSeq"][0], ['GDP396', 'ARG201'])
        _np.testing.assert_array_equal(site["pairs"]["AAresSeq"][1], ['GDP396', 'THR204'])
        _np.testing.assert_array_equal(site["pairs"]["AAresSeq"][2], ['GDP396', 'VAL202'])
        _np.testing.assert_array_equal(site["pairs"]["AAresSeq"][3], ['THR204', 'SER54'])
        _np.testing.assert_array_equal(site["pairs"]["AAresSeq"][4], ['MG397', 'THR204'])

    def test_runs_w_residx(self):
        site = mdciao.sites.x2site({"name": "interesting contacts",
                                    "pairs": {"residx": [
                                        "353-972",
                                        "340-956",
                                        "343-956",
                                        "344-956",
                                        "340-959",
                                        "343-865"
                                    ]}})
        self.assertDictEqual(site, {"pairs": {"residx": [
            [353,972],
            [340,956],
            [343,956],
            [344,956],
            [340,959],
            [343,865]
        ]}, "name": "interesting contacts",
        "n_pairs":6}
                             )

    def test_runs_w_residx_no_str(self):
        site = mdciao.sites.x2site({"pairs":
            {"residx": [
                [353, 972],
                [340, 956],
                [343, 956],
                [344, 956],
                [340, 959],
                [343, 865]]},
            "name": "interesting contacts",
            "n_pairs": 6})
        #it's the same dict repeated, nothing should change

        self.assertDictEqual(site,{"pairs":
            {"residx": [
                [353, 972],
                [340, 956],
                [343, 956],
                [344, 956],
                [340, 959],
                [343, 865]]},
            "name": "interesting contacts",
            "n_pairs": 6})

    def test_runs_w_residx_str_list(self):
        site = mdciao.sites.x2site({"pairs":
            {"residx": [
                ["353", "972"],
                ["340", "956"],
                ["343", "956"],
                ["344", "956"],
                ["340", "959"],
                ["343", "865"]]},
            "name": "interesting contacts",
            "n_pairs": 6})

        print(site)

    def test_raises(self):
        with self.assertRaises(KeyError):
             mdciao.sites.x2site({"fonds":
                {"residx": [
                    [353, 972],
                    [340, 956],
                    [343, 956],
                    [344, 956],
                    [340, 959],
                    [343, 865]]},
                "name": "interesting contacts",
                "n_pairs": 6})


    def test_runs_w_dict(self):
        site = mdciao.sites.x2site(test_filenames.GDP_json)
        _np.testing.assert_equal(site["name"], "GDP")
        _np.testing.assert_equal(site["n_pairs"], 5)
        _np.testing.assert_equal(len(site["pairs"]["AAresSeq"]), 5)
        _np.testing.assert_array_equal(site["pairs"]["AAresSeq"][0], ['GDP396', 'ARG201'])
        _np.testing.assert_array_equal(site["pairs"]["AAresSeq"][1], ['GDP396', 'THR204'])
        _np.testing.assert_array_equal(site["pairs"]["AAresSeq"][2], ['GDP396', 'VAL202'])
        _np.testing.assert_array_equal(site["pairs"]["AAresSeq"][3], ['THR204', 'SER54'])
        _np.testing.assert_array_equal(site["pairs"]["AAresSeq"][4], ['MG397', 'THR204'])

    def test_runs_w_dict_of_lists_of_res(self):
        site = mdciao.sites.x2site(test_filenames.GDP_json)
        #Run it again, should yield the same dict
        self.assertDictEqual(site, mdciao.sites.x2site(site))


    def test_runs_w_residx_and_comment(self):
        site = mdciao.sites.x2site({"name": "interesting contacts",
                                    "pairs": {"residx": [
                                        "353-972",
                                        "340-956",
                                        "# 343-956",
                                        "344-956",
                                        "340-959",
                                        "343-865"
                                    ]}})
        self.assertDictEqual(site, {"pairs": {"residx": [
            [353, 972],
            [340, 956],
            # [343,956],
            [344, 956],
            [340, 959],
            [343, 865]
        ]}, "name": "interesting contacts",
            "n_pairs": 5})

    def test_raises_wo_name(self):
        site = mdciao.sites.x2site(test_filenames.GDP_name_json)
        site.pop("name")
        with _np.testing.assert_raises(ValueError):
            mdciao.sites.x2site(site)

    def test_runs_with_name(self):
        site = mdciao.sites.x2site(test_filenames.GDP_name_json)
        _np.testing.assert_equal(site["name"], "siteGDP")

    def test_runs_with_dat(self):
        site = mdciao.sites.x2site(test_filenames.tip_dat)
        site_test = mdciao.sites.x2site(test_filenames.tip_json)
        self.assertDictEqual(site["pairs"], site_test["pairs"])
        self.assertEqual(site["name"],"tip.json as plain ascii")
        self.assertEqual(site["n_pairs"],site_test["n_pairs"])

class Test_dat2site(unittest.TestCase):

    def test_works_AAresSeq(self):
        site = siteIO.dat2site(test_filenames.tip_dat)
        self.assertDictEqual(site, {"pairs": {"AAresSeq": [
            "L394-K270",
            "D381-Q229",
            "Q384-Q229",
            "R385-Q229",
            "D381-K232",
            "Q384-I135"
        ]}, "name":"tip.json as plain ascii"}
                             )
    def test_works_residx(self):
        site = siteIO.dat2site(test_filenames.tip_residx_dat,fmt="residx")
        print(site)
        self.assertDictEqual(site, {"pairs": {"residx": [
            "353-972",
            "340-956",
            "343-956",
            "344-956",
            "340-959",
            "343-865"
        ]}, "name": "tip.json as plain ascii with residx"}
                             )
    def test_works_consensus(self):
        site = siteIO.dat2site(test_filenames.tip_consensus_dat,fmt="consensus")
        print(site)
        self.assertDictEqual(site, {"pairs": {"consensus": [
            "G.H5.26-6.32x32",
            "G.H5.13-5.68x68",
            "G.H5.16-5.68x68",
            "G.H5.17-5.68x68",
            "G.H5.13-5.71x71",
            "G.H5.16-3.54x54"
        ]}, "name": "tip.json as plain ascii with consensus"}
                             )
    def test_Valuerror(self):
        with self.assertRaises(ValueError):
            siteIO.dat2site(test_filenames.tip_residx_dat, fmt="AAresSeq")
        with self.assertRaises(ValueError):
            siteIO.dat2site(test_filenames.tip_dat, fmt="residx")


class Test_site2str(unittest.TestCase):

    def test_sitefile(self):
        mdciao.sites.site2str(test_filenames.GDP_json)
    def test_sitedict(self):
        site = mdciao.sites.x2site(test_filenames.GDP_json)
        mdciao.sites.site2str(site)
    def test_what(self):
        with _np.testing.assert_raises(ValueError):
            mdciao.sites.site2str([1])

class Test_sites_to_res_pairs(unittest.TestCase):
    def setUp(self):
        self.GDP_json = test_filenames.GDP_json
        self.geom = _md.load(test_filenames.actor_pdb)
        self.fragments = mdciao.fragments.get_fragments(self.geom.top)
        self.GPCR_map = mdciao.examples.GPCRLabeler_ardb2_human().top2labels(self.geom.top)
        self.CGN_map = mdciao.examples.CGNLabeler_gnas2_human().top2labels(self.geom.top)
    def test_the_idxs_work_no_frags(self):
        site = mdciao.sites.x2site(self.GDP_json)
        ctc_idxs, __ = mdciao.sites.sites_to_res_pairs([site], self.geom.top)
        for (ii,jj), (resi,resj) in zip(ctc_idxs,site["pairs"]["AAresSeq"]):
            _np.testing.assert_equal(str(self.geom.top.residue(ii)),resi)
            _np.testing.assert_equal(str(self.geom.top.residue(jj)),resj)

    def test_the_idxs_work_w_frags(self):
        site = mdciao.sites.x2site(self.GDP_json)
        ctc_idxs, __ = mdciao.sites.sites_to_res_pairs([site], self.geom.top,
                                                       fragments=self.fragments)
        for (ii,jj), (resi,resj) in zip(ctc_idxs,site["pairs"]["AAresSeq"]):
            _np.testing.assert_equal(str(self.geom.top.residue(ii)),resi)
            _np.testing.assert_equal(str(self.geom.top.residue(jj)),resj)

    def test_Nones_get_differentiated_and_added(self):
        # Even though the first two ones yield both (None,None) pairs, they
        # will get added to the list. The third one will be recognized
        # as a seen None,None
        ctc_idxs, site_maps = mdciao.sites.sites_to_res_pairs([{"name": "bogus1",       "pairs": {"AAresSeq": ["AAA1-AAA2"]}},
                                                               {"name": "bogus2",       "pairs": {"AAresSeq": ["AAA3-AAA4"]}},
                                                               {"name": "bogus1_repeat","pairs": {"AAresSeq": ["AAA1-AAA2"]}},
                                                               ], self.geom.top)
        _np.testing.assert_array_equal(ctc_idxs, [[None,None],
                                                  [None,None]])
        self.assertListEqual(site_maps, [[0],[1],[0]])

    def test_consensus(self):
        ctc_idxs, site_maps = mdciao.sites.sites_to_res_pairs([{"name": "interesting contacts",
                                                                "pairs": {"consensus": [
                                                                    "G.H5.26-6.32x32"]}}], self.geom.top,
                                                              consensus_maps={"GPCR": self.GPCR_map,
                                                                              "CGN": self.CGN_map})
        _np.testing.assert_array_equal(ctc_idxs, [[659, 209]])
        _np.testing.assert_array_equal(site_maps, [[0]])
    def test_consensus_fails_no_consensus_passed(self):
        with self.assertRaises(ValueError):
            mdciao.sites.sites_to_res_pairs([{"name":"interesting contacts",
                                              "pairs": {"consensus": [
                                                  "G.H5.26-6.32x32"]}}], self.geom.top)

    def test_consensus_no_label_present(self):
        ctc_idxs, site_maps = mdciao.sites.sites_to_res_pairs([{"name":"interesting contacts",
                                              "pairs": {"consensus": [
                                                  "G.H5.26-6.32x32"]}}], self.geom.top,
                                            consensus_maps={"GPCR":self.GPCR_map})
        _np.testing.assert_array_equal(ctc_idxs, [[None, 209]])
        _np.testing.assert_array_equal(site_maps, [[0]])

class Test_discard_empty_sites(unittest.TestCase):

    def setUp(self):
        self.site_list = [
            {'name': 'site0', 'pairs': {'AAresSeq': ['ALA20-ALA21',  # ALA20 doesn't exist
                                                     'GLU101-GLU122']}},  # both exist

            {'name': 'site1', 'pairs': {'AAresSeq': ['GLU31-ALA20',  # GLU31 and ALA20 don't exist
                                                     'GLU17-GLU12']}},  # both exist

            {'name': 'site2', 'pairs': {'AAresSeq': ['GLN101-ALA122']}},  # GLN101 doesn't exist

            {'name': 'site3', 'pairs': {'AAresSeq': ['GLU101-GLU122']}}]  # both exist, but was seen before
        self.top = _md.load(test_filenames.top_pdb).top

        self.ctc_idxs, self.site_maps = mdciao.sites.sites_to_res_pairs(self.site_list, self.top)

        # This has to work otherwise the below tests don't make sense
        _np.testing.assert_array_equal(self.ctc_idxs, [[None, 374],
                                                  [69, 852],
                                                  [None, None],
                                                  [709, 365],
                                                  [None, None]])
        self.assertListEqual(self.site_maps, [[0, 1],
                                              [2, 3],
                                              [4],
                                              [1]])

    def test_just_works(self):

        new_ctc_idxs, new_site_maps, new_sites, discarded = mdciao.sites.discard_empty_sites(self.ctc_idxs, self.site_maps, self.site_list)
        _np.testing.assert_array_equal(new_ctc_idxs, [
                                                # [None, 374],
                                                [69, 852],
                                                # [None, None],
                                                [709, 365],
                                                # [None, None]
                                                ])
        self.assertListEqual(new_site_maps, [[0],
                                         [1],
                                         [0],
                                         ])
        self.assertListEqual(discarded["partial"], [0,1])
        self.assertListEqual(discarded["full"],[2])
        assert len(new_site_maps)==3
        self.assertDictEqual(new_sites[0], {"name": "site0", "pairs": {"residx": [[69, 852]]}, "n_pairs": 1})
        self.assertDictEqual(new_sites[1], {"name": "site1", "pairs": {"residx": [[709, 365]]}, "n_pairs": 1})
        self.assertDictEqual(new_sites[2], {"name": "site3", "pairs": {"residx": [[69, 852]]}, "n_pairs": 1})

    def test_just_works_full(self):
        new_ctc_idxs, new_site_maps, new_sites, discarded = mdciao.sites.discard_empty_sites(self.ctc_idxs, self.site_maps,
                                                                                             self.site_list, allow_partial_sites=False)
        _np.testing.assert_array_equal(new_ctc_idxs, [
            # [None, 374],
            [69, 852],
            # [None, None],
            #[709, 365],
            # [None, None]
        ])
        self.assertListEqual(new_site_maps, [[0]])
        self.assertListEqual(discarded["partial"], [])
        self.assertListEqual(discarded["full"], [0,1,2])
        assert len(new_site_maps)==1
        self.assertDictEqual(new_sites[0], {"name": "site3", "pairs": {"residx": [[69, 852]]}, "n_pairs": 1})

    def test_all_empty(self):
        site_list = self.site_list[:2]  # The first two sites have each one nonexistent pair
        ctc_idxs, site_maps = mdciao.sites.sites_to_res_pairs(site_list, self.top)
        new_ctc_idxs, new_site_maps, new_sites, discarded = mdciao.sites.discard_empty_sites(ctc_idxs,
                                               site_maps,
                                               site_list[:2],
                                               allow_partial_sites=False, #both sites will be discarded
                                               )
        assert len(new_ctc_idxs)==len(new_site_maps)==len(new_sites)==0
        assert len(discarded["partial"])==0
        assert discarded["full"]==[0,1]

if __name__ == '__main__':
    unittest.main()


import os
import io as _io
from contextlib import redirect_stdout
import unittest
from mdciao.examples import examples
from mdciao import contacts
from tempfile import mkdtemp, TemporaryDirectory
from unittest import mock
from glob import glob

import contextlib
@contextlib.contextmanager
def remember_cwd():
    curdir = os.getcwd()
    try:
        yield
    finally:
        os.chdir(curdir)

from mdciao.examples import filenames as test_filenames
class Test_ContactGroupL394(unittest.TestCase):

    def test_works(self):
        CG = examples.ContactGroupL394()
        assert isinstance(CG,contacts.ContactGroup)

    def test_except(self):
        with self.assertRaises(Exception):
            CG = examples.ContactGroupL394(bogus_arg="bogus")

class Test_Interface_B2AR_Gas(unittest.TestCase):
    def test_works(self):
        intf = examples.Interface_B2AR_Gas()
        assert intf.is_interface

    def test_except(self):
        with self.assertRaises(Exception):
            CG = examples.Interface_B2AR_Gas(bogus_arg="bogus")

class Test_KLIFSLabeler_P31751(unittest.TestCase):
    def test_works(self):
        KLIFS = examples.KLIFSLabeler_P31751()
        assert KLIFS._conlab_column == "KLIFS"

class Test_AlignerConsensus_B2AR_HUMAN_vs_OPSD_BOVIN(unittest.TestCase):
    def test_works(self):
        AC = examples.AlignerConsensus_B2AR_HUMAN_vs_OPSD_BOVIN()
        self.assertDictEqual({'B2AR': 'R131', 'OPS': 'R135'},
                             AC.AAresSeq_match("3.50x50")[["B2AR","OPS"]].squeeze().to_dict())

class Test_Zip(unittest.TestCase):

    def test_works(self):
        with TemporaryDirectory(suffix="_mdciao_test_zip") as td:
            zipfile = os.path.join(td,os.path.basename(test_filenames.zipfile_two_empties))
            os.symlink(test_filenames.zipfile_two_empties, zipfile)
            unzipped = examples._unzip2dir(zipfile)
            assert unzipped == os.path.splitext(zipfile)[0]
            os.path.exists(os.path.join("A.txt"))
            os.path.exists(os.path.join("B.txt"))

    def test_doesn_overwrite(self):
        with TemporaryDirectory(suffix="_mdciao_test_zip") as td:
            zipfile = os.path.join(td,os.path.basename(test_filenames.zipfile_two_empties))
            os.symlink(test_filenames.zipfile_two_empties, zipfile)
            unzipped = examples._unzip2dir(zipfile)
            b = _io.StringIO()
            with redirect_stdout(b) as stdout:
                unzipped = examples._unzip2dir(zipfile)
            stdout : _io.StringIO
            output = stdout.getvalue().splitlines()
            assert len(output)==3
            assert output[0]=="Unzipping to '%s/two_empty_files'"%td
            assert output[1]=="No unzipping of 'A.txt': file already exists."
            assert output[2]=="No unzipping of 'B.dat': file already exists."

class Test_recursive_funct(unittest.TestCase):
    def test_works_as_first(self):
        res = examples._recursive_prompt("this_file_doesnt_exist.txt","this_file_doesnt_exist")
        assert res =="this_file_doesnt_exist.txt"

    def test_has_to_iterate_once(self):
        with TemporaryDirectory(suffix="_mdciao_test_recursive") as td:
            with remember_cwd():
                os.chdir(td)
                open("this_file_exists.txt","w").close()
                input_values = (val for val in [""])
                with mock.patch('builtins.input', lambda *x: next(input_values)):
                    res = examples._recursive_prompt("this_file_exists.txt", "this_file_exists",is_file=True,verbose=True)
                    assert os.path.realpath(res)==os.path.join(os.path.realpath(td),"this_file_exists_01.txt")

    def test_has_to_enter_recursion(self):
        with TemporaryDirectory(suffix="_mdciao_test_recursive") as td:
            with remember_cwd():
                os.chdir(td)
                open("this_file_exists.txt","w").close()
                open("this_file_exists_01.txt","w").close()
                input_values = (val for val in ["this_file_exists_01.txt",""])
                with mock.patch('builtins.input', lambda *x: next(input_values)):
                    res = examples._recursive_prompt("this_file_exists.txt", "this_file_exists",is_file=True,verbose=True)
                    assert os.path.realpath(res)==os.path.join(os.path.realpath(td),"this_file_exists_03.txt")

    def test_new_file_is_good(self):
        with TemporaryDirectory(suffix="_mdciao_test_recursive") as td:
            with remember_cwd():
                os.chdir(td)
                open("this_file_exists.txt","w").close()
                open("this_file_exists_01.txt","w").close()
                input_values = (val for val in ["new_file_totally_different.txt",""])
                with mock.patch('builtins.input', lambda *x: next(input_values)):
                    res = examples._recursive_prompt("this_file_exists.txt", "this_file_exists",is_file=True,verbose=True)
                    assert os.path.realpath(res)==os.path.join(os.path.realpath(td),"new_file_totally_different.txt")

    def test_escapes_recursion(self):
        with TemporaryDirectory(suffix="_mdciao_test_recursive") as td:
            with remember_cwd():
                os.chdir(td)
                for ii in range(50):
                    open("test_%02u.dat" % ii, "w").close()
                with self.assertRaises(RecursionError):
                    examples._recursive_prompt("test_00.dat", "test", is_file=True)

    def test_skip_on_existing(self):
        with TemporaryDirectory(suffix="_mdciao_test_recursive") as td:
            with remember_cwd():
                os.chdir(td)
                open("test.00.dat", "w").close()
                examples._recursive_prompt("test_00.dat", "test", is_file=True, skip_on_existing=True)

class Test_down_safely(unittest.TestCase):

    def test_just_works(self):
        with TemporaryDirectory(suffix="_mdciao_test_down_safely") as td:
            with remember_cwd():
                os.chdir(td)
                local_path = examples._down_url_safely("https://proteinformatics.uni-leipzig.de/mdciao/mdciao_test_small.zip",verbose=True)
                assert os.path.exists(local_path)
    def test_different_name(self):
        with TemporaryDirectory(suffix="_mdciao_test_down_safely") as td:
            with remember_cwd():
                os.chdir(td)
                local_path = examples._down_url_safely("https://proteinformatics.uni-leipzig.de/mdciao/mdciao_test_small.zip",verbose=True,
                                                       rename_to="myfile.zip")
                assert local_path.endswith("myfile.zip")
                assert os.path.exists(local_path)

    def test_skip_on_existing(self):
        with TemporaryDirectory(suffix="_mdciao_test_down_safely") as td:
            with remember_cwd():
                os.chdir(td)
                with open("mdciao_test_small.zip","w") as f:
                    f.write("Won't be overwrriten")
                local_path = examples._down_url_safely("https://proteinformatics.uni-leipzig.de/mdciao/mdciao_test_small.zip",
                                                       verbose=True, skip_on_existing=True)
                assert open("mdciao_test_small.zip").read() == "Won't be overwrriten"
                assert local_path.endswith("mdciao_test_small.zip")
                assert os.path.exists(local_path)

class Test_fetch_example_data(unittest.TestCase):

    def test_just_works(self):
        with TemporaryDirectory(suffix="_mdciao_test_fetch") as td:
            with remember_cwd():
                os.chdir(td)
                local_path = examples.fetch_example_data("https://proteinformatics.uni-leipzig.de/mdciao/mdciao_test_small.zip",unzip=False)
                assert os.path.exists(local_path)
                #assert os.path.exists((os.path.splitext(local_path))[0])
                files =  os.listdir(td)
                assert len(files)==1
                assert files[0]=="mdciao_test_small.zip"

    def test_just_unzip(self):
        with TemporaryDirectory(suffix="_mdciao_test_fetch") as td:
            with remember_cwd():
                os.chdir(td)
                local_path = examples.fetch_example_data("https://proteinformatics.uni-leipzig.de/mdciao/mdciao_test_small.zip",unzip=True)
                assert os.path.exists(local_path)
                assert os.path.exists((os.path.splitext(local_path))[0])
                files =  sorted(os.listdir(td))
                assert len(files)==2
                assert files[0]=="mdciao_test_small"
                assert files[1]=="mdciao_test_small.zip"
                unzipped_files =sorted(os.listdir(files[0]))
                assert len(unzipped_files)==2
                assert unzipped_files[0]=="A.dat"
                assert unzipped_files[1]=="B.dat"

    def test_alias(self):
        with TemporaryDirectory(suffix="_mdciao_test_fetch") as td:
            with remember_cwd():
                os.chdir(td)
                local_path = examples.fetch_example_data("test",
                                                         unzip=False)
                assert os.path.exists(local_path)
                # assert os.path.exists((os.path.splitext(local_path))[0])
                files = os.listdir(td)
                assert len(files) == 1
                assert files[0] == "mdciao_test_small.zip"


    def test_alias_unzip_to_otherfile(self):
        with TemporaryDirectory(suffix="_mdciao_test_fetch") as td:
            with remember_cwd():
                os.chdir(td)
                local_path = examples.fetch_example_data("test",
                                                         unzip="unzip_here")
                assert os.path.exists(local_path)
                # assert os.path.exists((os.path.splitext(local_path))[0])
                files = os.listdir(td)
                assert len(files) == 2
                assert files[0] == "unzip_here.zip"
                assert files[1] == "unzip_here"
                extracted = sorted(os.listdir(files[1]))
                assert extracted[0] == "A.dat"
                assert extracted[1] == "B.dat"


    def test_skip_on_existing(self):
        with TemporaryDirectory(suffix="_mdciao_test_fetch") as td:
            with remember_cwd():
                os.chdir(td)
                local_path = examples.fetch_example_data("test",
                                                         unzip=False)
                assert os.path.exists(local_path)
                # assert os.path.exists((os.path.splitext(local_path))[0])
                files = os.listdir(td)
                assert len(files) == 1
                assert files[0] == "mdciao_test_small.zip"
                # Create a fake file to test it doesn't ovewrite
                with open("mdciao_test_small.zip", "w") as f:
                    f.write("Won't be overwrriten")
                local_path = examples.fetch_example_data("test",
                                                         unzip=False, skip_on_existing=True)
                assert open("mdciao_test_small.zip").read() == "Won't be overwrriten"


class Test_notebooks(unittest.TestCase):

    def test_just_works(self):
        with TemporaryDirectory(suffix="_mdciao_test_notebooks") as td:
            new_notebook_path = examples.notebooks(os.path.join(td,"mdciao_notebooks"))
            assert os.path.realpath(new_notebook_path) != os.path.realpath(test_filenames.notebooks_path)
            new_nbs = sorted([os.path.basename(ff) for ff in os.listdir(new_notebook_path)])
            old_nbs = sorted([os.path.basename(ff) for ff in glob(os.path.join(test_filenames.notebooks_path,"*ipynb"))])
            self.assertListEqual(new_nbs, old_nbs)


class Test_Filenames(unittest.TestCase):

    def test_works(self):
        for attr in dir(examples.filenames):
            if not attr.startswith("_"):
                assert os.path.exists(getattr(examples.filenames,attr))

if __name__ == '__main__':
    unittest.main()
